import sys
from typing import List

import atheris
from fuzz_helpers import EnhancedFuzzedDataProvider

with atheris.instrument_imports():
    import dateparser

import re

import pytz

import dateparser.data
import dateparser.parser

language_codes = dateparser.data.languages_info.language_order
directives = [
    "%a",
    "%A",
    "%w",
    "%d",
    "%b",
    "%B",
    "%m",
    "%y",
    "%Y",
    "%H",
    "%I",
    "%p",
    "%M",
    "%S",
    "%f",
    "%z",
    "%Z",
    "%j",
    "%U",
    "%W",
    "%c",
    "%x",
    "%X",
    "%%",
    "%G",
    "%u",
    "%V",
    "%:Z",
]
locale_codes = ["fr-PF", "qu-EC", "af-NA"]
date_order = list(dateparser.parser.date_order_chart.keys())
timezone = list(pytz.all_timezones)
preferred_date = ["last", "first", "current"]
preferred_dates_from = ["past", "future", "current_period"]
parsers = [
    "timestamp",
    "negative-timestamp",
    "relative-time",
    "custom-formats",
    "absolute-time",
    "no-spaces-time",
]


def _get_format_strings(fdp: EnhancedFuzzedDataProvider) -> List[str]:
    format_strings = []
    for _ in range(fdp.ConsumeIntInRange(0, 5)):
        format_strings.append(fdp.ConsumeString(1).join(fdp.ConsumeSublist(directives)))
    return format_strings


def TestOneInput(data):
    fdp = EnhancedFuzzedDataProvider(data)

    settings = {
        "DATE_ORDER": fdp.PickValueInList(date_order),
        "PREFER_LOCALE_DATE_ORDER": fdp.ConsumeBool(),
        "TIMEZONE": fdp.PickValueInList(timezone),
        "TO_TIMEZONE": fdp.PickValueInList(timezone),
        "RETURN_AS_TIMEZONE_AWARE": fdp.ConsumeBool(),
        "PREFER_MONTH_OF_YEAR": fdp.PickValueInList(preferred_date),
        "PREFER_DAY_OF_MONTH": fdp.PickValueInList(preferred_date),
        "PREFER_DATES_FROM": fdp.PickValueInList(preferred_dates_from),
        "RELATIVE_BASE": fdp.ConsumeDate(),
        "STRICT_PARSING": fdp.ConsumeBool(),
        "REQUIRE_PARTS": [],
        "SKIP_TOKENS": [
            fdp.ConsumeRandomString() for _ in range(fdp.ConsumeIntInRange(0, 3))
        ],
        "NORMALIZE": fdp.ConsumeBool(),
        "RETURN_TIME_AS_PERIOD": fdp.ConsumeBool(),
        "PARSERS": fdp.ConsumeSublist(parsers),
        "DEFAULT_LANGUAGES": fdp.ConsumeSublist(language_codes),
        "LANGUAGE_DETECTION_CONFIDENCE_THRESHOLD": fdp.ConsumeProbability(),
    }

    try:
        dateparser.parse(
            fdp.ConsumeRandomString(),
            date_formats=_get_format_strings(fdp),
            languages=fdp.ConsumeSublist(language_codes),
            locales=fdp.ConsumeSublist(locale_codes),
            region=fdp.ConsumeString(2),
            settings=settings,
        )
    except re.error:
        return -1


def main():
    atheris.Setup(sys.argv, TestOneInput)
    atheris.Fuzz()


if __name__ == "__main__":
    main()



from datetime import datetime

from parameterized import param, parameterized

from dateparser.calendars.jalali import JalaliCalendar
from dateparser.calendars.jalali_parser import PersianDate, jalali_parser
from tests import BaseTestCase


class TestPersianDate(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.persian_date = NotImplemented

    def when_date_is_given(self, year, month, day):
        self.persian_date = PersianDate(year, month, day)

    def then_weekday_is(self, weekday):
        self.assertEqual(self.persian_date.weekday(), weekday)

    @parameterized.expand(
        [
            param(year=1348, month=1, day=3, weekday=0),
            param(year=1348, month=2, day=28, weekday=0),
            param(year=1348, month=3, day=27, weekday=2),
            param(year=1348, month=4, day=11, weekday=3),
        ]
    )
    def test_weekday(self, year, month, day, weekday):
        self.when_date_is_given(year, month, day)
        self.then_weekday_is(weekday)


class TestJalaliParser(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.translated = NotImplemented

    def when_date_is_given(self, date_string):
        self.translated = jalali_parser.to_latin(date_string)

    def then_month_is_parsed_as(self, month_name):
        self.assertEqual(month_name, self.translated)

    def then_weekday_is_parsed_as(self, weekday):
        self.assertEqual(weekday, self.translated)

    def then_digit_is_parsed_as(self, digit):
        self.assertEqual(digit, self.translated)

    def then_numeral_parsed_is(self, datetime):
        self.assertEqual(datetime, self.translated)

    @parameterized.expand(
        [
            param(date_string="فروردین", latin="Farvardin"),
            param(date_string="اردیبهشت", latin="Ordibehesht"),
            param(date_string="خرداد", latin="Khordad"),
            param(date_string="تیر", latin="Tir"),
            param(date_string="مرداد", latin="Mordad"),
            param(date_string="شهریور", latin="Shahrivar"),
            param(date_string="مهر", latin="Mehr"),
            param(date_string="آبان", latin="Aban"),
            param(date_string="آذر", latin="Azar"),
            param(date_string="دی", latin="Dey"),
            param(date_string="بهمن", latin="Bahman"),
            param(date_string="اسفند", latin="Esfand"),
        ]
    )
    def test_replace_months(self, date_string, latin):
        self.when_date_is_given(date_string)
        self.then_month_is_parsed_as(latin)

    @parameterized.expand(
        [
            param(date_string="یکشنبه", latin="Sunday"),
            param(date_string="روز شنبه", latin="Saturday"),
            param(date_string="دوشنبه", latin="Monday"),
            param(date_string="سهشنبه", latin="Tuesday"),
            param(date_string="چهار شنبه", latin="Wednesday"),
            param(date_string="پنج شنبه", latin="Thursday"),
            param(date_string="جمعه", latin="Friday"),
        ]
    )
    def test_replace_weekdays(self, date_string, latin):
        self.when_date_is_given(date_string)
        self.then_weekday_is_parsed_as(latin)

    @parameterized.expand(
        [
            param(date_string="۰۱", latin="01"),
            param(date_string="۲۵", latin="25"),
            param(date_string="۶۷۴", latin="674"),
            param(date_string="۰۵۶۱۲۷۳۸۹۴", latin="0561273894"),
        ]
    )
    def test_replace_digits(self, date_string, latin):
        self.when_date_is_given(date_string)
        self.then_digit_is_parsed_as(latin)

    @parameterized.expand(
        [
            param(persian="صفر", numeral="0"),
            param(persian="یک", numeral="1"),
            param(persian="اول", numeral="1"),
            param(persian="دو", numeral="2"),
            param(persian="دوم", numeral="2"),
            param(persian="سه", numeral="3"),
            param(persian="سوم", numeral="3"),
            param(persian="چهار", numeral="4"),
            param(persian="چهارم", numeral="4"),
            param(persian="پنج", numeral="5"),
            param(persian="پنجم", numeral="5"),
            param(persian="شش", numeral="6"),
            param(persian="ششم", numeral="6"),
            param(persian="هفت", numeral="7"),
            param(persian="هفتم", numeral="7"),
            param(persian="هشت", numeral="8"),
            param(persian="هشتمین", numeral="8"),
            param(persian="نه", numeral="9"),
            param(persian="نهم", numeral="9"),
            param(persian="ده", numeral="10"),
            param(persian="دهم", numeral="10"),
            param(persian="یازده", numeral="11"),
            param(persian="یازدهم", numeral="11"),
            param(persian="دوازده", numeral="12"),
            param(persian="دوازدهم", numeral="12"),
            param(persian="سیزده", numeral="13"),
            param(persian="سیزدهم", numeral="13"),
            param(persian="چهارده", numeral="14"),
            param(persian="چهاردهم", numeral="14"),
            param(persian="پانزده", numeral="15"),
            param(persian="پانزدهمین", numeral="15"),
            param(persian="شانزده", numeral="16"),
            param(persian="شانزدهم", numeral="16"),
            param(persian="هفده", numeral="17"),
            param(persian="هفدهم", numeral="17"),
            param(persian="هجده", numeral="18"),
            param(persian="هجدهم", numeral="18"),
            param(persian="نوزده", numeral="19"),
            param(persian="نوزدهم", numeral="19"),
            param(persian="بیست", numeral="20"),
            param(persian="بیستم", numeral="20"),
            param(persian="بیست و یک", numeral="21"),
            param(persian="بیست و دو", numeral="22"),
            param(persian="بیست ثانیه", numeral="22"),
            param(persian="بیست و سه", numeral="23"),
            param(persian="بیست و سوم", numeral="23"),
            param(persian="بیست و چهار", numeral="24"),
            param(persian="بیست و چهارم", numeral="24"),
            param(persian="بیست و پنج", numeral="25"),
            param(persian="بیست و شش", numeral="26"),
            param(persian="بیست و هفت", numeral="27"),
            param(persian="بیست و هفتمین", numeral="27"),
            param(persian="بیست و هشت", numeral="28"),
            param(persian="بیست و نه", numeral="29"),
            param(persian="سی", numeral="30"),
            param(persian="سی و یک", numeral="31"),
            param(persian="سی و یکم", numeral="31"),
        ]
    )
    def test_replace_day_literal_with_numerals(self, persian, numeral):
        self.when_date_is_given(persian)
        self.then_numeral_parsed_is(numeral)


class TestJalaliCalendar(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.result = NotImplemented
        self.date_string = NotImplemented
        self.calendar = NotImplemented

    def when_date_is_given(self, date_string):
        self.date_string = date_string
        self.calendar = JalaliCalendar(date_string)
        self.result = self.calendar.get_date()

    def then_date_parsed_is(self, datetime):
        self.assertEqual(datetime, self.result["date_obj"])

    @parameterized.expand(
        [
            param(
                date_string="سه شنبه سوم شهریور ۱۳۹۴", dt=datetime(2015, 8, 25, 0, 0)
            ),
            param(date_string="پنجشنبه چهارم تیر ۱۳۹۴", dt=datetime(2015, 6, 25, 0, 0)),
            param(date_string="شنبه یکم فروردین ۱۳۹۴", dt=datetime(2015, 3, 21, 0, 0)),
            param(
                date_string="یکشنبه شانزدهم آذر ۱۳۹۳", dt=datetime(2014, 12, 7, 0, 0)
            ),
            param(
                date_string="دوشنبه دوازدهم آبان ۱۳۹۳", dt=datetime(2014, 11, 3, 0, 0)
            ),
            param(date_string="یکشنبه سیزدهم مهر ۱۳۹۳", dt=datetime(2014, 10, 5, 0, 0)),
            param(date_string="دوشنبه دوم تیر ۱۳۹۳", dt=datetime(2014, 6, 23, 0, 0)),
            param(
                date_string="یکشنبه سوم فروردین ۱۳۹۳", dt=datetime(2014, 3, 23, 0, 0)
            ),
            param(date_string="پنجشنبه دهم بهمن ۱۳۹۲", dt=datetime(2014, 1, 30, 0, 0)),
            param(date_string="شنبه چهاردهم دی ۱۳۹۲", dt=datetime(2014, 1, 4, 0, 0)),
            param(date_string="جمعه بیستم خرداد ۱۳۹۰", dt=datetime(2011, 6, 10, 0, 0)),
            param(date_string="شنبه نهم مرداد ۱۳۸۹", dt=datetime(2010, 7, 31, 0, 0)),
            param(
                date_string="پنجشنبه بیست و سوم اردیبهشت ۱۳۸۹",
                dt=datetime(2010, 5, 13, 0, 0),
            ),
            param(date_string="جمعه سی ام اسفند ۱۳۸۷", dt=datetime(2009, 3, 20, 0, 0)),
            # dates with time component
            param(
                date_string="پنجشنبه ۲۶ شهريور ۱۳۹۴ ساعت ۹:۳۲",
                dt=datetime(2015, 9, 17, 9, 32),
            ),
            param(
                date_string="دوشنبه ۲۳ شهريور ۱۳۹۴ ساعت ۱۹:۱۱",
                dt=datetime(2015, 9, 14, 19, 11),
            ),
            param(
                date_string="جمعه سی ام اسفند ۱۳۸۷ساعت 19:47",
                dt=datetime(2009, 3, 20, 19, 47),
            ),
            param(
                date_string="شنبه چهاردهم دی ۱۳۹۲ساعت 12:1",
                dt=datetime(2014, 1, 4, 12, 1),
            ),
            param(
                date_string="پنجشنبه 26 شهریور 1394 ساعت ساعت 11 و 01 دقیقه و 47 ثانیه",
                dt=datetime(2015, 9, 17, 11, 1, 47),
            ),
            param(
                date_string="پنجشنبه 26 شهریور 1394 ساعت ساعت 10 و 58 دقیقه و 04 ثانیه",
                dt=datetime(2015, 9, 17, 10, 58, 4),
            ),
            param(
                date_string="سه شنبه 17 شهریور 1394 ساعت ساعت 18 و 21 دقیقه و 44 ثانیه",
                dt=datetime(2015, 9, 8, 18, 21, 44),
            ),
            param(date_string="پنجشنبه 11 تیر 1394", dt=datetime(2015, 7, 2, 0, 0)),
            param(
                date_string="پنجشنبه 26 شهریور 1394 ساعت ساعت 11 و 01 دقیقه",
                dt=datetime(2015, 9, 17, 11, 1),
            ),
        ]
    )
    def test_get_date(self, date_string, dt):
        self.when_date_is_given(date_string)
        self.then_date_parsed_is(dt)



from unittest import TestCase


class BaseTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.__patches = []

        self.error = NotImplemented

    def add_patch(self, patch):
        patch.start()
        self.__patches.append(patch)

    def tearDown(self):
        super().tearDown()
        for patch in reversed(self.__patches):
            patch.stop()

    def then_error_was_raised(self, error_cls, allowed_substrings=()):
        self.assertIsInstance(self.error, error_cls)
        self.assertTrue(
            any(mesg in str(self.error) for mesg in allowed_substrings),
            "Didn't found any of the expected messages (%r) -- message was: %r"
            % (allowed_substrings, self.error),
        )



#!/usr/bin/env python
#
# dateparser documentation build configuration file, created by
# sphinx-quickstart on Tue Jul  9 22:26:36 2013.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# file.


import os
import sys

# Get the project root dir, which is the parent dir of this
cwd = os.getcwd()
project_root = os.path.dirname(cwd)

# Insert the project root dir as the first element in the PYTHONPATH.
# This lets us ensure that the source package is imported, and that its
# version is used.
sys.path.insert(0, project_root)

import dateparser

# -- General configuration ---------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_rtd_theme",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "DateParser"
copyright = "2014, Scrapinghub"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = dateparser.__version__
# The full version, including alpha/beta/rc tags.
release = dateparser.__version__

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets)
# here, relative to this directory. They are copied after the builtin
# static files, so a file named "default.css" will overwrite the builtin
# "default.css".
html_static_path = []

# Output file base name for HTML help builder.
htmlhelp_basename = "dateparserdoc"


# -- Options for LaTeX output ------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    ("index", "dateparser.tex", "DateParser Documentation", "Scrapinghub", "manual"),
]


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("index", "dateparser", "DateParser Documentation", ["Scrapinghub"], 1)]


# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "dateparser",
        "DateParser Documentation",
        "Scrapinghub",
        "dateparser",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# sphinx.ext.intersphinx confs
intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}


html_theme_options = {
    "logo_only": True,
    "collapse_navigation": True,
    "sticky_navigation": True,
    "navigation_depth": 4,
    "includehidden": True,
    "titles_only": False,
}

html_logo = "../artwork/dateparser-logo.png"



import unittest
from datetime import datetime
from unittest.mock import Mock

from parameterized import param, parameterized

from dateparser import parse
from dateparser.custom_language_detection.fasttext import (
    detect_languages as fast_text_detect_languages,
)
from dateparser.custom_language_detection.langdetect import (
    detect_languages as lang_detect_detect_languages,
)
from dateparser.date import DateDataParser
from dateparser.search import search_dates

detect_languages = Mock()
detect_languages.return_value = ["en"]


class CustomLangDetectParserTest(unittest.TestCase):
    def check_is_returned_list(self):
        self.assertEqual(type(self.result), list)

    @parameterized.expand(
        [
            param(dt_string="14 June 2020", confidence_threshold=0.0),
            param(dt_string="26 July 2021", confidence_threshold=0.0),
        ]
    )
    def test_custom_language_detect_fast_text(self, dt_string, confidence_threshold):
        self.result = fast_text_detect_languages(dt_string, confidence_threshold)
        self.check_is_returned_list()

    @parameterized.expand(
        [
            param(dt_string="14 June 2020", confidence_threshold=0.0),
        ]
    )
    def test_custom_language_detect_lang_detect(self, dt_string, confidence_threshold):
        self.result = lang_detect_detect_languages(dt_string, confidence_threshold)
        self.check_is_returned_list()

    @parameterized.expand(
        [
            param(dt_string="10-10-2021", confidence_threshold=0.5),
        ]
    )
    def test_lang_detect_doesnt_raise_error(self, dt_string, confidence_threshold):
        result = lang_detect_detect_languages(dt_string, confidence_threshold)
        assert result == []

    # Mock test for parse, search_dates and DateDataParser

    detect_languages = Mock()
    detect_languages.return_value = ["en"]

    # parse

    def when_date_is_parsed_using_parse(self, dt_string):
        self.result = parse(dt_string, detect_languages_function=detect_languages)

    def then_date_obj_exactly_is(self, expected_date_obj):
        self.assertEqual(expected_date_obj, self.result)

    @parameterized.expand(
        [
            param("Tuesday Jul 22, 2014", datetime(2014, 7, 22, 0, 0, 0)),
        ]
    )
    def test_custom_language_detect_mock_parse(self, dt_string, expected_date_obj):
        self.when_date_is_parsed_using_parse(dt_string)
        self.then_date_obj_exactly_is(expected_date_obj)

    # DateDataParser

    def when_date_is_parsed_using_with_datedataparser(self, dt_string):
        ddp = DateDataParser(detect_languages_function=detect_languages)
        self.result = ddp.get_date_data(dt_string)["date_obj"]

    @parameterized.expand(
        [
            param("Tuesday Jul 22, 2014", datetime(2014, 7, 22, 0, 0, 0)),
        ]
    )
    def test_custom_language_detect_mock_datedataparser(
        self, dt_string, expected_date_obj
    ):
        self.when_date_is_parsed_using_with_datedataparser(dt_string)
        self.then_date_obj_exactly_is(expected_date_obj)

    # search_date

    def when_date_is_parsed_using_with_search_dates(self, dt_string):
        self.result = search_dates(
            dt_string, detect_languages_function=detect_languages
        )

    @parameterized.expand(
        [
            param(
                "January 3, 2017 - February 1st",
                [
                    ("January 3, 2017", datetime(2017, 1, 3, 0, 0)),
                    ("February 1st", datetime(2017, 2, 1, 0, 0)),
                ],
            ),
        ]
    )
    def test_custom_language_detect_mock_search_dates(
        self, dt_string, expected_date_obj
    ):
        self.when_date_is_parsed_using_with_search_dates(dt_string)
        self.then_date_obj_exactly_is(expected_date_obj)



from operator import attrgetter

import regex as re
from parameterized import param, parameterized

from dateparser.languages.loader import default_loader
from tests import BaseTestCase


class TestLoading(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.locale_generator = NotImplemented

    @classmethod
    def setUpClass(cls):
        cls.data_loader = default_loader
        cls.data_loader._loaded_locales = {}
        cls.data_loader._loaded_languages = {}

    @parameterized.expand(
        [
            param(
                given_languages=["es", "fr", "en"],
                given_locales=None,
                given_region=None,
                loaded_languages=["en", "es", "fr"],
                loaded_locales=["en", "es", "fr"],
                expected_locales=["es", "fr", "en"],
            ),
            param(
                given_languages=None,
                given_locales=["fr-BF", "ar-SO", "asa"],
                given_region=None,
                loaded_languages=["en", "es", "fr", "ar", "asa"],
                loaded_locales=["en", "es", "fr", "fr-BF", "ar-SO", "asa"],
                expected_locales=["fr-BF", "ar-SO", "asa"],
            ),
            param(
                given_languages=None,
                given_locales=["nl-CW", "so-KE", "sr-Latn-XK", "zgh"],
                given_region=None,
                loaded_languages=[
                    "en",
                    "es",
                    "fr",
                    "ar",
                    "asa",
                    "nl",
                    "so",
                    "sr-Latn",
                    "zgh",
                ],
                loaded_locales=[
                    "en",
                    "es",
                    "fr",
                    "fr-BF",
                    "ar-SO",
                    "asa",
                    "nl-CW",
                    "so-KE",
                    "sr-Latn-XK",
                    "zgh",
                ],
                expected_locales=["nl-CW", "so-KE", "sr-Latn-XK", "zgh"],
            ),
            param(
                given_languages=["pt", "zh-Hant", "zh-Hans"],
                given_locales=None,
                given_region="MO",
                loaded_languages=[
                    "en",
                    "es",
                    "fr",
                    "ar",
                    "asa",
                    "nl",
                    "so",
                    "sr-Latn",
                    "zgh",
                    "pt",
                    "zh-Hant",
                    "zh-Hans",
                ],
                loaded_locales=[
                    "en",
                    "es",
                    "fr",
                    "fr-BF",
                    "ar-SO",
                    "asa",
                    "nl-CW",
                    "so-KE",
                    "sr-Latn-XK",
                    "zgh",
                    "pt-MO",
                    "zh-Hant-MO",
                    "zh-Hans-MO",
                ],
                expected_locales=["pt-MO", "zh-Hant-MO", "zh-Hans-MO"],
            ),
            param(
                given_languages=["pt", "he"],
                given_locales=["ru-UA", "ckb-IR", "sq-XK"],
                given_region="MO",
                loaded_languages=[
                    "en",
                    "es",
                    "fr",
                    "ar",
                    "asa",
                    "nl",
                    "so",
                    "sr-Latn",
                    "zgh",
                    "pt",
                    "zh-Hant",
                    "zh-Hans",
                    "ru",
                    "ckb",
                    "sq",
                ],
                loaded_locales=[
                    "en",
                    "es",
                    "fr",
                    "fr-BF",
                    "ar-SO",
                    "asa",
                    "nl-CW",
                    "so-KE",
                    "sr-Latn-XK",
                    "zgh",
                    "pt-MO",
                    "zh-Hant-MO",
                    "zh-Hans-MO",
                    "ru-UA",
                    "ckb-IR",
                    "sq-XK",
                ],
                expected_locales=["ru-UA", "ckb-IR", "sq-XK"],
            ),
            param(
                given_languages=["da", "ja"],
                given_locales=["shi-Latn", "teo-KE", "ewo", "vun"],
                given_region="AO",
                loaded_languages=[
                    "en",
                    "es",
                    "fr",
                    "ar",
                    "asa",
                    "nl",
                    "so",
                    "sr-Latn",
                    "zgh",
                    "pt",
                    "zh-Hant",
                    "zh-Hans",
                    "ru",
                    "ckb",
                    "sq",
                    "shi-Latn",
                    "teo",
                    "ewo",
                    "vun",
                ],
                loaded_locales=[
                    "en",
                    "es",
                    "fr",
                    "fr-BF",
                    "ar-SO",
                    "asa",
                    "nl-CW",
                    "so-KE",
                    "sr-Latn-XK",
                    "zgh",
                    "pt-MO",
                    "zh-Hant-MO",
                    "zh-Hans-MO",
                    "ru-UA",
                    "ckb-IR",
                    "sq-XK",
                    "shi-Latn",
                    "teo-KE",
                    "ewo",
                    "vun",
                ],
                expected_locales=["shi-Latn", "teo-KE", "ewo", "vun"],
            ),
        ]
    )
    def test_loading(
        self,
        given_languages,
        given_locales,
        given_region,
        loaded_languages,
        loaded_locales,
        expected_locales,
    ):
        self.load_data(
            languages=given_languages, locales=given_locales, region=given_region
        )
        self.then_locales_are_yielded_in_order(expected_locales)
        self.then_loaded_languages_are(loaded_languages)
        self.then_loaded_locales_are(loaded_locales)

    def load_data(self, languages, locales, region):
        self.locale_generator = self.data_loader.get_locales(
            languages=languages, locales=locales, region=region, use_given_order=True
        )

    def then_loaded_languages_are(self, loaded_languages):
        self.assertCountEqual(
            loaded_languages, self.data_loader._loaded_languages.keys()
        )

    def then_loaded_locales_are(self, loaded_locales):
        self.assertCountEqual(loaded_locales, self.data_loader._loaded_locales.keys())

    def then_locales_are_yielded_in_order(self, expected_locales):
        self.assertEqual(
            list(map(attrgetter("shortname"), list(self.locale_generator))),
            expected_locales,
        )


class TestLocaleDataLoader(BaseTestCase):
    UNKNOWN_LANGUAGES_EXCEPTION_RE = re.compile(r"Unknown language\(s\): (.+)")
    UNKNOWN_LOCALES_EXCEPTION_RE = re.compile(r"Unknown locale\(s\): (.+)")

    def setUp(self):
        super().setUp()
        self.data_loader = default_loader
        self.data_loader._loaded_languages = {}
        self.data_loader._loaded_locales = {}

    @parameterized.expand(
        [
            param(given_locales=["es-MX", "ar-EG", "fr-DJ"]),
            param(given_locales=["pt-MO", "ru-KZ", "es-CU"]),
            param(given_locales=["zh-Hans-HK", "en-VG", "my"]),
            param(given_locales=["tl", "nl-SX", "de-BE"]),
        ]
    )
    def test_loading_with_given_order(self, given_locales):
        self.load_data(given_locales, use_given_order=True)
        self.then_locales_are_yielded_in_order(given_locales)

    @parameterized.expand(
        [
            param(
                given_locales=["os-RU", "ln-CF", "ee-TG"],
                expected_locales=["ee-TG", "ln-CF", "os-RU"],
            ),
            param(given_locales=["fo-DK", "khq"], expected_locales=["khq", "fo-DK"]),
            param(
                given_locales=["en-CC", "fr-BE", "ar-KW"],
                expected_locales=["en-CC", "fr-BE", "ar-KW"],
            ),
        ]
    )
    def test_loading_without_given_order(self, given_locales, expected_locales):
        self.load_data(given_locales, use_given_order=False)
        self.then_locales_are_yielded_in_order(expected_locales)

    @parameterized.expand(
        [
            param(given_locales=["sw-KE", "ru-UA", "he"]),
            param(given_locales=["de-IT", "ta-MY", "pa-Arab"]),
            param(given_locales=["bn-IN", "pt-ST", "ko-KP", "ta"]),
            param(given_locales=["fr-NE", "ar-SY"]),
        ]
    )
    def test_get_locale_map_with_given_order(self, given_locales):
        self.given_locale_map(locales=given_locales, use_given_order=True)
        self.then_locale_map_in_order(given_locales)

    @parameterized.expand(
        [
            param(
                given_locales=["en-FJ", "pt-CV", "fr-RW"],
                expected_locales=["en-FJ", "fr-RW", "pt-CV"],
            ),
            param(
                given_locales=["pt-AO", "hi", "zh-Hans-SG", "vi"],
                expected_locales=["pt-AO", "zh-Hans-SG", "vi", "hi"],
            ),
            param(
                given_locales=["gsw-FR", "es-BZ", "ca-IT", "qu-EC"],
                expected_locales=["es-BZ", "ca-IT", "qu-EC", "gsw-FR"],
            ),
        ]
    )
    def test_get_locale_map_without_given_order(self, given_locales, expected_locales):
        self.given_locale_map(locales=given_locales, use_given_order=False)
        self.then_locale_map_in_order(expected_locales)

    @parameterized.expand(
        [
            param(
                given_languages=["es", "ar-001", "xx"],
                unknown_languages=["ar-001", "xx"],
            ),
            param(
                given_languages=["sr-Latn", "sq", "ii-Latn"],
                unknown_languages=["ii-Latn"],
            ),
            param(given_languages=["vol", "bc"], unknown_languages=["vol", "bc"]),
        ]
    )
    def test_error_raised_for_unknown_languages(
        self, given_languages, unknown_languages
    ):
        self.given_locale_map(languages=given_languages)
        self.then_error_for_unknown_languages_raised(unknown_languages)

    @parameterized.expand(
        [
            param(
                given_locales=["es-AB", "ar-001", "fr-DJ"],
                unknown_locales=["es-AB", "ar-001"],
            ),
            param(
                given_locales=["ru-MD", "my-MY", "zz"], unknown_locales=["my-MY", "zz"]
            ),
            param(given_locales=["nl-SX", "be-BE", "ca-FR"], unknown_locales=["be-BE"]),
        ]
    )
    def test_error_raised_for_unknown_locales(self, given_locales, unknown_locales):
        self.given_locale_map(locales=given_locales)
        self.then_error_for_unknown_locales_raised(unknown_locales)

    @parameterized.expand(
        [
            param(given_locales=["en-TK", "en-TO", "zh"]),
            param(given_locales=["es-PY", "es-IC", "ja", "es-DO"]),
            param(given_locales=["ca-TA", "ca", "fr"]),
        ]
    )
    def test_error_raised_for_conflicting_locales(self, given_locales):
        self.given_locale_map(locales=given_locales)
        self.then_error_was_raised(
            ValueError, "Locales should not have same language and different region"
        )

    @parameterized.expand(
        [
            param(given_locales=["en-TK", "en-TO", "zh"]),
            param(given_locales=["es-PY", "es-IC", "ja", "es-DO"]),
            param(given_locales=["af-NA", "da", "af"]),
        ]
    )
    def test_conflicting_locales_load_if_allow_conflicting_locales(self, given_locales):
        self.load_data(
            given_locales, use_given_order=True, allow_conflicting_locales=True
        )
        self.then_locales_are_yielded_in_order(given_locales)

    def load_data(
        self, given_locales, use_given_order=False, allow_conflicting_locales=False
    ):
        self.locale_generator = self.data_loader.get_locales(
            locales=given_locales,
            use_given_order=use_given_order,
            allow_conflicting_locales=allow_conflicting_locales,
        )

    def given_locale_map(self, languages=None, locales=None, use_given_order=True):
        try:
            self.locale_map = self.data_loader.get_locale_map(
                languages=languages, locales=locales, use_given_order=use_given_order
            )
        except Exception as error:
            self.error = error

    def then_locales_are_yielded_in_order(self, expected_locales):
        self.assertEqual(
            list(map(attrgetter("shortname"), list(self.locale_generator))),
            expected_locales,
        )

    def then_locale_map_in_order(self, expected_locales):
        self.assertEqual(list(self.locale_map.keys()), expected_locales)

    def then_error_for_unknown_languages_raised(self, unknown_languages):
        self.assertIsInstance(self.error, ValueError)
        match = self.UNKNOWN_LANGUAGES_EXCEPTION_RE.match(str(self.error))
        self.assertTrue(match)
        languages = match.group(1).split(", ")
        self.assertCountEqual(
            languages,
            [repr(unknown_language) for unknown_language in unknown_languages],
        )

    def then_error_for_unknown_locales_raised(self, unknown_locales):
        self.assertIsInstance(self.error, ValueError)
        match = self.UNKNOWN_LOCALES_EXCEPTION_RE.match(str(self.error))
        self.assertTrue(match)
        locales = match.group(1).split(", ")
        self.assertCountEqual(
            locales, [repr(unknown_locale) for unknown_locale in unknown_locales]
        )



from datetime import datetime

import pytest

from dateparser.date import DateData


class TestDateData:
    def test_get_item_like_dict(self):
        date = datetime(year=5432, month=3, day=1)
        dd = DateData(date_obj=date, period="day", locale="de")
        assert dd["date_obj"] == date
        assert dd["period"] == "day"
        assert dd["locale"] == "de"

    def test_get_item_like_dict_keyerror(self):
        dd = DateData(date_obj=None, period="day", locale="de")
        with pytest.raises(KeyError) as e:
            date_obj = dd["date"]
            assert e == "date"
            assert not date_obj

    def test_set_item_like_dict(self):
        dd = DateData()
        assert dd.date_obj is None

        date = datetime(year=5432, month=3, day=1)
        dd["date_obj"] = date
        assert dd.date_obj == date

    def test_set_item_like_dict_keyerror(self):
        dd = DateData()
        with pytest.raises(KeyError) as e:
            dd["date"] = datetime(year=5432, month=3, day=1)
            assert e == "date"

    @pytest.mark.parametrize(
        "date,period,locale,expected",
        [
            (
                datetime(year=2020, month=10, day=28),
                "day",
                "en",
                "DateData(date_obj=datetime.datetime(2020, 10, 28, 0, 0), period='day', locale='en')",
            ),
            (
                datetime(year=2014, month=5, day=29, hour=3, minute=2),
                "day",
                "es",
                "DateData(date_obj=datetime.datetime(2014, 5, 29, 3, 2), period='day', locale='es')",
            ),
            (
                datetime(
                    year=2028,
                    month=7,
                    day=31,
                    hour=3,
                    minute=2,
                    second=12,
                    microsecond=601265,
                ),
                "time",
                "fr",
                "DateData(date_obj=datetime.datetime(2028, 7, 31, 3, 2, 12, 601265), period='time', locale='fr')",
            ),
            (
                datetime(year=1994, month=8, day=1),
                "month",
                "ca",
                "DateData(date_obj=datetime.datetime(1994, 8, 1, 0, 0), period='month', locale='ca')",
            ),
            (
                datetime(year=2033, month=10, day=12),
                "month",
                None,
                "DateData(date_obj=datetime.datetime(2033, 10, 12, 0, 0), period='month', locale=None)",
            ),
            (None, "day", None, "DateData(date_obj=None, period='day', locale=None)"),
            (None, "year", "fr", "DateData(date_obj=None, period='year', locale='fr')"),
        ],
    )
    def test_repr(self, date, period, locale, expected):
        dd = DateData(date_obj=date, period=period, locale=locale)
        assert dd.__repr__() == expected



from dateparser_scripts.write_complete_data import write_complete_data


def test_dateparser_data_integrity():
    files = write_complete_data(in_memory=True)

    for filename in files:
        with open(filename, "rb") as f:
            assert f.read().strip() == files[filename].strip(), (
                'The content of the file "{}" doesn\'t match the content'
                " of the generated file.".format(filename)
            )



from datetime import date, datetime

from parameterized import param, parameterized
from pytz import utc

import dateparser
from tests import BaseTestCase


class TestParseFunction(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.result = NotImplemented

    @parameterized.expand(
        [
            param(date_string="24 de Janeiro de 2014", expected_date=date(2014, 1, 24)),
            param(date_string="2 de Enero de 2013", expected_date=date(2013, 1, 2)),
            param(date_string="January 25, 2014", expected_date=date(2014, 1, 25)),
        ]
    )
    def test_parse_dates_in_different_languages(self, date_string, expected_date):
        self.when_date_is_parsed_with_defaults(date_string)
        self.then_parsed_date_is(expected_date)

    @parameterized.expand(
        [
            param(
                date_string="May 5, 2000 13:00",
                expected_date=datetime(2000, 5, 5, 13, 0),
            ),
            param(
                date_string="August 8, 2018 5 PM",
                expected_date=datetime(2018, 8, 8, 17, 0),
            ),
            param(
                date_string="February 26, 1981 5 am UTC",
                expected_date=datetime(1981, 2, 26, 5, 0, tzinfo=utc),
            ),
        ]
    )
    def test_parse_dates_with_specific_time(self, date_string, expected_date):
        self.when_date_is_parsed_with_defaults(date_string)
        self.then_parsed_date_and_time_is(expected_date)

    @parameterized.expand(
        [
            param(
                date_string="May 5, 2000 13:00",
                expected_date=datetime(2000, 5, 5, 13, 0),
                relative=datetime(2000, 1, 1, 0, 0, tzinfo=utc),
            ),
            param(
                date_string="August 8, 2018 5 PM",
                expected_date=datetime(2018, 8, 8, 17, 0),
                relative=datetime(1900, 5, 5, 0, 0, tzinfo=utc),
            ),
            param(
                date_string="February 26, 1981 5 am UTC",
                expected_date=datetime(1981, 2, 26, 5, 0, tzinfo=utc),
                relative=datetime(1981, 2, 26, 5, 0, tzinfo=utc),
            ),
        ]
    )
    def test_parse_dates_with_specific_time_and_settings(
        self, date_string, expected_date, relative
    ):
        self.when_date_is_parsed_with_settings(
            date_string, settings={"RELATIVE_BASE": relative}
        )
        self.then_parsed_date_and_time_is(expected_date)

    @parameterized.expand(
        [
            param(
                date_string="24 de Janeiro de 2014",
                languages=["pt"],
                expected_date=date(2014, 1, 24),
            ),
        ]
    )
    def test_dates_which_match_languages_are_parsed(
        self, date_string, languages, expected_date
    ):
        self.when_date_is_parsed(date_string, languages=languages)
        self.then_parsed_date_is(expected_date)

    @parameterized.expand(
        [
            param(date_string="January 24, 2014", languages=["pt"]),
        ]
    )
    def test_dates_which_do_not_match_languages_are_not_parsed(
        self, date_string, languages
    ):
        self.when_date_is_parsed(date_string, languages=languages)
        self.then_date_was_not_parsed()

    @parameterized.expand(
        [
            param(
                date_string="24 de Janeiro de 2014",
                locales=["pt-TL"],
                expected_date=date(2014, 1, 24),
            ),
        ]
    )
    def test_dates_which_match_locales_are_parsed(
        self, date_string, locales, expected_date
    ):
        self.when_date_is_parsed(date_string, locales=locales)
        self.then_parsed_date_is(expected_date)

    @parameterized.expand(
        [
            param(date_string="January 24, 2014", locales=["pt-AO"]),
        ]
    )
    def test_dates_which_do_not_match_locales_are_not_parsed(
        self, date_string, locales
    ):
        self.when_date_is_parsed(date_string, locales=locales)
        self.then_date_was_not_parsed()

    def when_date_is_parsed_with_defaults(self, date_string):
        self.result = dateparser.parse(date_string)

    def when_date_is_parsed(self, date_string, languages=None, locales=None):
        self.result = dateparser.parse(
            date_string, languages=languages, locales=locales
        )

    def when_date_is_parsed_with_settings(self, date_string, settings=None):
        self.result = dateparser.parse(date_string, settings=settings)

    def then_parsed_date_is(self, expected_date):
        self.assertEqual(
            self.result, datetime.combine(expected_date, datetime.min.time())
        )

    def then_parsed_date_and_time_is(self, expected_date):
        self.assertEqual(self.result, expected_date)

    def then_date_was_not_parsed(self):
        self.assertIsNone(self.result)



import itertools
from datetime import datetime

import pytest
from parameterized import param, parameterized
from pytz import UnknownTimeZoneError, utc

from dateparser.conf import settings
from dateparser.utils import (
    apply_timezone,
    apply_timezone_from_settings,
    find_date_separator,
    get_last_day_of_month,
    get_next_leap_year,
    get_previous_leap_year,
    localize_timezone,
    registry,
)
from tests import BaseTestCase


class TestUtils(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.date_format = None
        self.result = None

    def given_date_format(self, date_format):
        self.date_format = date_format

    def when_date_separator_is_parsed(self):
        self.result = find_date_separator(self.date_format)

    def then_date_separator_is(self, sep):
        self.assertEqual(self.result, sep)

    @staticmethod
    def make_class_without_get_keys():
        class SomeClass:
            pass

        some_class = SomeClass
        return some_class

    @parameterized.expand(
        [
            param(date_format=fmt.format(sep=sep), expected_sep=sep)
            for (fmt, sep) in itertools.product(
                ["%d{sep}%m{sep}%Y", "%d{sep}%m{sep}%Y %H:%M"], ["/", ".", "-", ":"]
            )
        ]
    )
    def test_separator_extraction(self, date_format, expected_sep):
        self.given_date_format(date_format)
        self.when_date_separator_is_parsed()
        self.then_date_separator_is(expected_sep)

    @parameterized.expand(
        [
            param(datetime(2015, 12, 12), timezone="UTC", zone="UTC"),
            param(datetime(2015, 12, 12), timezone="Asia/Karachi", zone="Asia/Karachi"),
            param(datetime(2015, 12, 12, tzinfo=utc), timezone="UTC", zone="UTC"),
        ]
    )
    def test_localize_timezone_function(self, date, timezone, zone):
        tzaware_dt = localize_timezone(date, timezone)
        self.assertEqual(tzaware_dt.tzinfo.zone, zone)

    @parameterized.expand(
        [
            param(datetime(2015, 12, 12), timezone="UTB"),
            param(datetime(2015, 12, 12), timezone="Asia/Karach"),
        ]
    )
    def test_localize_timezone_function_raise_error(self, date, timezone):
        self.assertRaises(UnknownTimeZoneError, localize_timezone, date, timezone)

    @parameterized.expand(
        [
            param(datetime(2015, 12, 12), timezone="UTC+3", zone=r"UTC\+03:00"),
        ]
    )
    def test_localize_timezone_function_exception(self, date, timezone, zone):
        tzaware_dt = localize_timezone(date, timezone)
        self.assertEqual(tzaware_dt.tzinfo._StaticTzInfo__name, zone)

    @parameterized.expand(
        [
            param(
                datetime(2015, 12, 12, 10, 12),
                timezone="Asia/Karachi",
                expected=datetime(2015, 12, 12, 15, 12),
            ),
            param(
                datetime(2015, 12, 12, 10, 12),
                timezone="-0500",
                expected=datetime(2015, 12, 12, 5, 12),
            ),
        ]
    )
    def test_apply_timezone_function(self, date, timezone, expected):
        result = apply_timezone(date, timezone)
        result = result.replace(tzinfo=None)
        self.assertEqual(expected, result)

    @parameterized.expand(
        [
            param(
                datetime(2015, 12, 12, 10, 12),
                timezone="Asia/Karachi",
                expected=datetime(2015, 12, 12, 15, 12),
            ),
            param(
                datetime(2015, 12, 12, 10, 12),
                timezone="-0500",
                expected=datetime(2015, 12, 12, 5, 12),
            ),
        ]
    )
    def test_apply_timezone_from_settings_function(self, date, timezone, expected):
        result = apply_timezone_from_settings(
            date, settings.replace(**{"TO_TIMEZONE": timezone, "TIMEZONE": "UTC"})
        )
        self.assertEqual(expected, result)

    @parameterized.expand(
        [
            param(
                datetime(2015, 12, 12, 10, 12), expected=datetime(2015, 12, 12, 10, 12)
            ),
        ]
    )
    def test_apply_timezone_from_settings_function_none_settings(self, date, expected):
        result = apply_timezone_from_settings(date, None)
        self.assertEqual(expected, result)

    @parameterized.expand(
        [
            param(
                datetime(2015, 12, 12, 10, 12),
            ),
            param(
                datetime(2015, 12, 12, 10, 12),
            ),
        ]
    )
    def test_apply_timezone_from_settings_function_should_return_tz(self, date):
        result = apply_timezone_from_settings(
            date, settings.replace(**{"RETURN_AS_TIMEZONE_AWARE": True})
        )
        self.assertTrue(bool(result.tzinfo))

    def test_registry_when_get_keys_not_implemented(self):
        cl = self.make_class_without_get_keys()
        self.assertRaises(NotImplementedError, registry, cl)

    @parameterized.expand(
        [
            param(2111, 1, 31),
            param(1999, 2, 28),  # normal year
            param(1996, 2, 29),  # leap and not centurial year
            param(2000, 2, 29),  # leap and centurial year
            param(1700, 2, 28),  # no leap and centurial year (exception)
            param(2020, 3, 31),
            param(1987, 4, 30),
            param(1000, 5, 31),
            param(1534, 6, 30),
            param(1777, 7, 31),
            param(1234, 8, 31),
            param(1678, 9, 30),
            param(1947, 10, 31),
            param(2015, 11, 30),
            param(2300, 12, 31),
        ]
    )
    def test_get_last_day_of_month(self, year, month, expected_last_day):
        assert get_last_day_of_month(year, month) == expected_last_day


@pytest.mark.parametrize(
    "year,expected_previous_leap_year",
    [
        (2020, 2016),
        (2000, 1996),  # leap and centurial year
        (2104, 2096),  # missing no leap centurial year
        (1704, 1696),
        (2396, 2392),
        (0, -4),  # even if this is not a valid year, it is the expected result
    ],
)
def test_get_previous_leap_year(year, expected_previous_leap_year):
    assert get_previous_leap_year(year) == expected_previous_leap_year


@pytest.mark.parametrize(
    "year,expected_next_leap_year",
    [
        (2020, 2024),
        (1996, 2000),  # leap and centurial year
        (2096, 2104),  # missing no leap centurial year
        (1696, 1704),
        (2396, 2400),
        (0, 4),
    ],
)
def test_get_next_leap_year(year, expected_next_leap_year):
    assert get_next_leap_year(year) == expected_next_leap_year



import sys
import unittest
from datetime import datetime

from parameterized import param, parameterized

from dateparser.calendars.hijri import HijriCalendar
from tests import BaseTestCase


class TestHijriParser(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.result = NotImplemented
        self.date_string = NotImplemented
        self.parser = NotImplemented
        self.translated = NotImplemented

    def when_date_is_given(self, dt_string, date_formats, languages):
        self.date_string = dt_string
        self.parser = HijriCalendar(dt_string)
        self.result = self.parser.get_date()

    def then_parsed_datetime_is(self, dt):
        self.assertEqual(dt, self.result["date_obj"])

    @parameterized.expand(
        [
            param(
                dt_string="14-09-1432 هـ, 09:40 صباحاً",
                dt_obj=datetime(2011, 8, 14, 9, 40),
            ),
            param(
                dt_string="20-02-1430 هـ, 07:21 صباحاً",
                dt_obj=datetime(2009, 2, 15, 7, 21),
            ),
            param(
                dt_string="11-08-1434 هـ, 09:38 صباحاً",
                dt_obj=datetime(2013, 6, 20, 9, 38),
            ),
            param(
                dt_string=" 17-01-1437 هـ 08:30 مساءً",
                dt_obj=datetime(2015, 10, 30, 20, 30),
            ),
            param(
                dt_string="29-02-1433 هـ, 06:22 صباحاً",
                dt_obj=datetime(2012, 1, 23, 6, 22),
            ),
            param(dt_string="30-02-1433", dt_obj=datetime(2012, 1, 24)),
            param(
                dt_string="04-03-1433 هـ, 10:08 مساءً",
                dt_obj=datetime(2012, 1, 27, 22, 8),
            ),
        ]
    )
    def test_datetime_parsing(
        self, dt_string, dt_obj, date_formats=None, languages=None
    ):
        from dateparser.conf import settings

        settings.DATE_ORDER = "DMY"
        self.when_date_is_given(dt_string, date_formats, languages)
        self.then_parsed_datetime_is(dt_obj)
        settings.DATE_ORDER = "MDY"



import datetime as dt
from datetime import datetime, timedelta
from unittest.mock import Mock, patch

from parameterized import param, parameterized
from pytz import timezone

import dateparser.timezone_parser
from dateparser import parse
from dateparser.timezone_parser import (
    StaticTzInfo,
    get_local_tz_offset,
    pop_tz_offset_from_string,
)
from tests import BaseTestCase


class TestTZPopping(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.initial_string = (
            self.datetime_string
        ) = self.timezone_offset = NotImplemented

    @parameterized.expand(
        [
            param("Sep 03 2014 | 4:32 pm EDT", -4),
            param("17th October, 2034 @ 01:08 am PDT", -7),
            param("17th October, 2034 @ 01:08 am (PDT)", -7),
            param("October 17, 2014 at 7:30 am PST", -8),
            param("20 Oct 2014 13:08 CET", +1),
            param("20 Oct 2014 13:08cet", +1),
            param("Nov 25 2014 | 10:17 pm EST", -5),
            param("Nov 25 2014 | 10:17 pm +0600", +6),
            param("Nov 25 2014 | 10:17 pm -0930", -9.5),
            param("20 Oct 2014 | 05:17 am -1200", -12),
            param("20 Oct 2014 | 05:17 am +0000", 0),
            param("20 Oct 2014 | 05:17 am -0000", 0),
            param("15 May 2004", None),
            param("Wed Aug 05 12:00:00 EDTERR 2015", None),
            param("Wed Aug 05 12:00:00 EDT 2015", -4),
            param("April 10, 2016 at 12:00:00 UTC", 0),
            param("April 10, 2016 at 12:00:00 MEZ", 1),
            param("April 10, 2016 at 12:00:00 MESZ", 2),
            param("April 10, 2016 at 12:00:00 GMT+2", 2),
            param("April 10, 2016 at 12:00:00 UTC+2:00", 2),
            param("April 10, 2016 at 12:00:00 GMT+02:00", 2),
            param("April 10, 2016 at 12:00:00 UTC+5:30", 5.5),
            param("April 10, 2016 at 12:00:00 GMT+05:30", 5.5),
            param("April 10, 2016 at 12:00:00 UTC-2", -2),
            param("April 10, 2016 at 12:00:00 GMT-2:00", -2),
            param("April 10, 2016 at 12:00:00 UTC-02:00", -2),
            param("April 10, 2016 at 12:00:00 GMT-9:30", -9.5),
            param("April 10, 2016 at 12:00:00 UTC-09:30", -9.5),
            param("Thu, 24 Nov 2016 16:03:00 UT", 0),
            param("Fri Sep 23 2016 10:34:51 GMT+0800 (CST)", 8),
            param("Fri Sep 23 2016 10:34:51 GMT+12", 12),
            param("Fri Sep 23 2016 10:34:51 UTC+13", 13),
            param("Fri Sep 23 2016 10:34:51 GMT+1245 (CST)", 12.75),
            param("Fri Sep 23 2016 10:34:51 UTC+1245", 12.75),
            param("2019-07-17T12:30:00.000-03:30", -3.5),
            param("2019-07-17T12:30:00.000-02:30", -2.5),
            param("16. srpna 2021 9:59:44 SELČ", 2),
            param("16. srpna 2021 9:59:44 SEČ", 1),
            param("16. srpna 2021 9:59:44 ZEČ", 0),
            param("16. srpna 2021 9:59:44 VEČ", 2),
        ]
    )
    def test_extracting_valid_offset(self, initial_string, expected_offset):
        self.given_string(initial_string)
        self.when_offset_popped_from_string()
        self.then_offset_is(expected_offset)

    @parameterized.expand(
        [
            param("Sep 03 2014 | 4:32 pm EDT", "Sep 03 2014 | 4:32 pm "),
            param(
                "17th October, 2034 @ 01:08 am PDT", "17th October, 2034 @ 01:08 am "
            ),
            param("October 17, 2014 at 7:30 am PST", "October 17, 2014 at 7:30 am "),
            param("20 Oct 2014 13:08 CET", "20 Oct 2014 13:08 "),
            param("20 Oct 2014 13:08cet", "20 Oct 2014 13:08"),
            param("Nov 25 2014 | 10:17 pm EST", "Nov 25 2014 | 10:17 pm "),
            param(
                "17th October, 2034 @ 01:08 am +0700", "17th October, 2034 @ 01:08 am "
            ),
            param("Sep 03 2014 4:32 pm +0630", "Sep 03 2014 4:32 pm "),
        ]
    )
    def test_timezone_deleted_from_string(self, initial_string, result_string):
        self.given_string(initial_string)
        self.when_offset_popped_from_string()
        self.then_string_modified_to(result_string)

    def test_string_not_changed_if_no_timezone(self):
        self.given_string("15 May 2004")
        self.when_offset_popped_from_string()
        self.then_string_modified_to("15 May 2004")

    def given_string(self, string_):
        self.initial_string = string_

    def when_offset_popped_from_string(self):
        self.datetime_string, timezone_offset = pop_tz_offset_from_string(
            self.initial_string
        )
        if timezone_offset:
            self.timezone_offset = timezone_offset.utcoffset("")
        else:
            self.timezone_offset = timezone_offset

    def then_string_modified_to(self, expected_string):
        self.assertEqual(expected_string, self.datetime_string)

    def then_offset_is(self, expected_offset):
        delta = (
            timedelta(hours=expected_offset) if expected_offset is not None else None
        )
        self.assertEqual(delta, self.timezone_offset)


class TestLocalTZOffset(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.timezone_offset = NotImplemented

    @parameterized.expand(
        [
            param(utc="2014-08-20 4:32", local="2014-08-20 8:32", offset=+4),
            param(utc="2052-01-02 11:07", local="2052-01-02 10:07", offset=-1),
            param(utc="2013-12-31 23:59", local="2014-01-01 00:29", offset=+0.5),
            param(utc="2011-07-14 11:59", local="2011-07-13 23:59", offset=-12),
            param(utc="2014-10-18 17:41", local="2014-10-18 17:41", offset=0),
        ]
    )
    def test_timezone_offset_calculation(self, utc, local, offset):
        self.given_time(utc, local)
        self.when_offset_popped_from_string()
        self.then_offset_is(offset)

    def when_offset_popped_from_string(self):
        self.timezone_offset = get_local_tz_offset()

    def then_offset_is(self, expected_offset):
        delta = (
            timedelta(seconds=3600 * expected_offset)
            if expected_offset is not None
            else None
        )
        self.assertEqual(delta, self.timezone_offset)

    def given_time(self, utc_dt_string, local_dt_string):
        datetime_cls = dateparser.timezone_parser.datetime
        if not isinstance(datetime_cls, Mock):
            datetime_cls = Mock(wraps=datetime)
        utc_dt_obj = datetime.strptime(utc_dt_string, "%Y-%m-%d %H:%M").astimezone(
            dt.timezone.utc
        )
        local_dt_obj = datetime.strptime(local_dt_string, "%Y-%m-%d %H:%M")

        def _dt_now(tz=None):
            if tz == dt.timezone.utc:
                return utc_dt_obj
            return local_dt_obj

        datetime_cls.now = Mock(side_effect=_dt_now)
        self.add_patch(patch("dateparser.timezone_parser.datetime", new=datetime_cls))


class TestTimeZoneConversion(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.settings = {}
        self.parser = parse
        self.result = NotImplemented

    @parameterized.expand(
        [
            param(
                "2015-12-31 10:04 AM",
                "Asia/Karachi",
                "UTC",
                datetime(2015, 12, 31, 5, 4),
            ),
            param(
                "2015-12-30 10:04 AM",
                "Asia/Karachi",
                "+0200",
                datetime(2015, 12, 30, 7, 4),
            ),
        ]
    )
    def test_timezone_conversion(self, datestring, from_tz, to_tz, expected):
        self.given_from_timezone(from_tz)
        self.given_to_timezone(to_tz)
        self.when_date_is_parsed(datestring)
        self.then_date_is(expected)

    def given_from_timezone(self, timezone):
        self.settings["TIMEZONE"] = timezone

    def given_to_timezone(self, timezone):
        self.settings["TO_TIMEZONE"] = timezone

    def when_date_is_parsed(self, datestring):
        self.result = self.parser(datestring, settings=self.settings)

    def then_date_is(self, date):
        self.assertEqual(date, self.result)


class TestStaticTzInfo(BaseTestCase):
    def setUp(self):
        super().setUp()

    @parameterized.expand(
        [
            param(given_date=datetime(2007, 1, 18, tzinfo=timezone("UTC"))),
            param(given_date=datetime(2003, 3, 31, tzinfo=timezone("US/Arizona"))),
            param(given_date=datetime(2000, 2, 20, tzinfo=timezone("Pacific/Samoa"))),
        ]
    )
    def test_localize_raises_error_if_date_has_tzinfo(self, given_date):
        self.timezone_info = StaticTzInfo("UTC\\+00:00", timedelta(0))
        self.when_date_is_localized(given_date)
        self.then_error_was_raised(
            ValueError, ["Not naive datetime (tzinfo is already set)"]
        )

    def when_date_is_localized(self, given_date):
        try:
            self.localized_date = self.timezone_info.localize(given_date)
        except Exception as error:
            self.error = error

    def then_localized_date_is(self, expected_date, expected_tzname):
        self.assertEqual(self.localized_date.date(), expected_date.date())
        self.assertEqual(self.localized_date.tzname(), expected_tzname)



from itertools import chain

import regex as re
from parameterized import parameterized

from dateparser.data import language_locale_dict
from dateparser.languages import default_loader
from tests import BaseTestCase

DEFAULT_MONTH_PATTERN = re.compile(r"^M?\d+$", re.U)
INVALID_AM_PM_PATTERN = re.compile(r"^[AaPp]\.?\s+[Mm]$")

all_locale_shortnames = list(
    chain(language_locale_dict.keys(), *language_locale_dict.values())
)
all_locales = list(
    default_loader.get_locales(
        locales=all_locale_shortnames, allow_conflicting_locales=True
    )
)
all_locale_params = [[locale] for locale in all_locales]

VALID_KEYS = [
    "name",
    "date_order",
    "skip",
    "pertain",
    "simplifications",
    "no_word_spacing",
    "ago",
    "in",
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
    "january",
    "february",
    "march",
    "april",
    "may",
    "june",
    "july",
    "august",
    "september",
    "october",
    "november",
    "december",
    "decade",
    "year",
    "month",
    "week",
    "day",
    "hour",
    "minute",
    "second",
    "am",
    "pm",
    "relative-type",
    "relative-type-regex",
    "sentence_splitter_group",
]

NECESSARY_KEYS = [
    "name",
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
    "january",
    "february",
    "march",
    "april",
    "may",
    "june",
    "july",
    "august",
    "september",
    "october",
    "november",
    "december",
    "year",
    "month",
    "week",
    "day",
    "hour",
    "minute",
    "second",
]

MONTH_NAMES = [
    "january",
    "february",
    "march",
    "april",
    "may",
    "june",
    "july",
    "august",
    "september",
    "october",
    "november",
    "december",
]


def is_invalid_translation(translation):
    return (not (translation and isinstance(translation, str))) or "." in translation


def is_invalid_month_translation(translation):
    return (
        (not (translation and isinstance(translation, str)))
        or "." in translation
        or DEFAULT_MONTH_PATTERN.match(translation)
    )


def is_invalid_am_pm_translation(translation):
    return (
        (not (translation and isinstance(translation, str)))
        or "." in translation
        or INVALID_AM_PM_PATTERN.match(translation)
    )


def is_invalid_simplification(simplification):
    if not isinstance(simplification, dict) or len(simplification) != 1:
        return True
    key, value = list(simplification.items())[0]
    return not isinstance(key, str) or not isinstance(value, str)


def is_invalid_relative_mapping(relative_mapping):
    key, value = relative_mapping
    if not (key and value and isinstance(key, str) and isinstance(value, list)):
        return True
    return not all([isinstance(x, str) for x in value])


def is_invalid_relative_regex_mapping(relative_regex_mapping):
    key, value = relative_regex_mapping
    if not (key and value and isinstance(key, str) and isinstance(value, list)):
        return True
    return not all([isinstance(x, str) for x in value])


class TestLocaleInfo(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.info = NotImplemented
        self.shortname = NotImplemented

    @parameterized.expand(all_locale_params)
    def test_extra_keys(self, locale):
        self.given_locale_info(locale)
        extra_keys = list(set(self.info.keys()) - set(VALID_KEYS))
        self.assertFalse(
            extra_keys, "Extra keys found for {}: {}".format(self.shortname, extra_keys)
        )

    @parameterized.expand(all_locale_params)
    def test_necessary_keys(self, locale):
        self.given_locale_info(locale)
        self.then_keys_present_in_info(NECESSARY_KEYS)
        self.then_translations_present_in_info(NECESSARY_KEYS)

    @parameterized.expand(all_locale_params)
    def test_name(self, locale):
        self.given_locale_info(locale)
        self.then_name_is_valid()

    @parameterized.expand(all_locale_params)
    def test_date_order(self, locale):
        self.given_locale_info(locale)
        self.then_date_order_is_valid()

    @parameterized.expand(all_locale_params)
    def test_january_translation(self, locale):
        self.given_locale_info(locale)
        self.then_month_translations_are_valid("january")

    @parameterized.expand(all_locale_params)
    def test_february_translation(self, locale):
        self.given_locale_info(locale)
        self.then_month_translations_are_valid("february")

    @parameterized.expand(all_locale_params)
    def test_march_translation(self, locale):
        self.given_locale_info(locale)
        self.then_month_translations_are_valid("march")

    @parameterized.expand(all_locale_params)
    def test_april_translation(self, locale):
        self.given_locale_info(locale)
        self.then_month_translations_are_valid("april")

    @parameterized.expand(all_locale_params)
    def test_may_translation(self, locale):
        self.given_locale_info(locale)
        self.then_month_translations_are_valid("may")

    @parameterized.expand(all_locale_params)
    def test_june_translation(self, locale):
        self.given_locale_info(locale)
        self.then_month_translations_are_valid("june")

    @parameterized.expand(all_locale_params)
    def test_july_translation(self, locale):
        self.given_locale_info(locale)
        self.then_month_translations_are_valid("july")

    @parameterized.expand(all_locale_params)
    def test_august_translation(self, locale):
        self.given_locale_info(locale)
        self.then_month_translations_are_valid("august")

    @parameterized.expand(all_locale_params)
    def test_september_translation(self, locale):
        self.given_locale_info(locale)
        self.then_month_translations_are_valid("september")

    @parameterized.expand(all_locale_params)
    def test_october_translation(self, locale):
        self.given_locale_info(locale)
        self.then_month_translations_are_valid("october")

    @parameterized.expand(all_locale_params)
    def test_november_translation(self, locale):
        self.given_locale_info(locale)
        self.then_month_translations_are_valid("november")

    @parameterized.expand(all_locale_params)
    def test_december_translation(self, locale):
        self.given_locale_info(locale)
        self.then_month_translations_are_valid("december")

    @parameterized.expand(all_locale_params)
    def test_am_translation(self, locale):
        self.given_locale_info(locale)
        self.then_am_pm_translations_are_valid("am")

    @parameterized.expand(all_locale_params)
    def test_pm_translation(self, locale):
        self.given_locale_info(locale)
        self.then_am_pm_translations_are_valid("pm")

    @parameterized.expand(all_locale_params)
    def test_sunday_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("sunday")

    @parameterized.expand(all_locale_params)
    def test_monday_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("monday")

    @parameterized.expand(all_locale_params)
    def test_tuesday_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("tuesday")

    @parameterized.expand(all_locale_params)
    def test_wednesday_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("wednesday")

    @parameterized.expand(all_locale_params)
    def test_thursday_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("thursday")

    @parameterized.expand(all_locale_params)
    def test_friday_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("friday")

    @parameterized.expand(all_locale_params)
    def test_saturday_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("saturday")

    @parameterized.expand(all_locale_params)
    def test_year_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("year")

    @parameterized.expand(all_locale_params)
    def test_month_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("month")

    @parameterized.expand(all_locale_params)
    def test_week_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("week")

    @parameterized.expand(all_locale_params)
    def test_day_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("day")

    @parameterized.expand(all_locale_params)
    def test_hour_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("hour")

    @parameterized.expand(all_locale_params)
    def test_minute_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("minute")

    @parameterized.expand(all_locale_params)
    def test_second_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("second")

    @parameterized.expand(all_locale_params)
    def test_ago_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("ago")

    @parameterized.expand(all_locale_params)
    def test_in_translation(self, locale):
        self.given_locale_info(locale)
        self.then_translations_are_valid("in")

    @parameterized.expand(all_locale_params)
    def test_skip_tokens(self, locale):
        self.given_locale_info(locale)
        self.then_skip_pertain_tokens_are_valid("skip")

    @parameterized.expand(all_locale_params)
    def test_pertain_tokens(self, locale):
        self.given_locale_info(locale)
        self.then_skip_pertain_tokens_are_valid("pertain")

    @parameterized.expand(all_locale_params)
    def test_simplifications(self, locale):
        self.given_locale_info(locale)
        self.then_simplifications_are_valid()

    @parameterized.expand(all_locale_params)
    def test_relative_type(self, locale):
        self.given_locale_info(locale)
        self.then_relative_type_is_valid()

    @parameterized.expand(all_locale_params)
    def test_relative_type_regex(self, locale):
        self.given_locale_info(locale)
        self.then_relative_type_regex_is_valid()

    @parameterized.expand(all_locale_params)
    def test_no_word_spacing(self, locale):
        self.given_locale_info(locale)
        self.then_no_word_spacing_is_valid()

    def given_locale_info(self, locale):
        self.info = locale.info
        self.shortname = locale.shortname

    def then_keys_present_in_info(self, keys_list):
        absent_keys = list(set(keys_list) - set(self.info.keys()))
        self.assertFalse(
            absent_keys,
            "{} not found in locale {}".format(", ".join(absent_keys), self.shortname),
        )

    def then_translations_present_in_info(self, keys_list):
        no_translation_keys = [key for key in keys_list if not self.info.get(key)]
        self.assertFalse(
            no_translation_keys,
            "Translations for {} not found in locale {}".format(
                ", ".join(no_translation_keys), self.shortname
            ),
        )

    def then_name_is_valid(self):
        name = self.info["name"]
        self.assertIsInstance(
            name,
            str,
            "Invalid type for name: {} for locale {}".format(
                type(name).__name__, self.shortname
            ),
        )
        self.assertEqual(
            name,
            self.shortname,
            "Invalid name: {} for locale {}".format(name, self.shortname),
        )

    def then_date_order_is_valid(self):
        if "date_order" in self.info:
            date_order = self.info["date_order"]
            self.assertIsInstance(
                date_order,
                str,
                "Invalid type for date_order: {} for locale {}".format(
                    type(date_order).__name__, self.shortname
                ),
            )
            self.assertIn(
                date_order,
                ["DMY", "DYM", "MDY", "MYD", "YDM", "YMD"],
                "Invalid date_order {} for {}".format(date_order, self.shortname),
            )

    def then_month_translations_are_valid(self, month):
        if month in self.info:
            month_translations = self.info[month]
            self.assertIsInstance(
                month_translations,
                list,
                "Invalid type for {}: {} for locale {}".format(
                    month, type(month_translations).__name__, self.shortname
                ),
            )
            invalid_translations = list(
                filter(is_invalid_month_translation, month_translations)
            )
            self.assertFalse(
                invalid_translations,
                "Invalid translations for {}: {} for locale {}".format(
                    month, ", ".join(map(repr, invalid_translations)), self.shortname
                ),
            )

    def then_am_pm_translations_are_valid(self, key):
        if key in self.info:
            translations_list = self.info[key]
            self.assertIsInstance(
                translations_list,
                list,
                "Invalid type for {}: {} for locale {}".format(
                    key, type(translations_list).__name__, self.shortname
                ),
            )
            invalid_translations = list(
                filter(is_invalid_am_pm_translation, translations_list)
            )
            self.assertFalse(
                invalid_translations,
                "Invalid translations for {}: {} for locale {}".format(
                    key, ", ".join(map(repr, invalid_translations)), self.shortname
                ),
            )

    def then_translations_are_valid(self, key):
        if key in self.info:
            translations_list = self.info[key]
            self.assertIsInstance(
                translations_list,
                list,
                "Invalid type for {}: {} for locale {}".format(
                    key, type(translations_list).__name__, self.shortname
                ),
            )
            invalid_translations = list(
                filter(is_invalid_translation, translations_list)
            )
            self.assertFalse(
                invalid_translations,
                "Invalid translations for {}: {} for locale {}".format(
                    key, ", ".join(map(repr, invalid_translations)), self.shortname
                ),
            )

    def then_skip_pertain_tokens_are_valid(self, key):
        if key in self.info:
            tokens_list = self.info[key]
            self.assertIsInstance(
                tokens_list,
                list,
                "Invalid type for {}: {} for locale {}".format(
                    key, type(tokens_list).__name__, self.shortname
                ),
            )
            invalid_tokens = [
                token
                for token in tokens_list
                if not token or not isinstance(token, str)
            ]
            self.assertFalse(
                invalid_tokens,
                "Invalid tokens for {}: {} for locale {}".format(
                    key, ", ".join(map(repr, invalid_tokens)), self.shortname
                ),
            )

    def then_simplifications_are_valid(self):
        if "simplifications" in self.info:
            simplifications_list = self.info["simplifications"]
            self.assertIsInstance(
                simplifications_list,
                list,
                "Invalid type for simplifications: {} for locale {}".format(
                    type(simplifications_list).__name__, self.shortname
                ),
            )
            invalid_simplifications = list(
                filter(is_invalid_simplification, simplifications_list)
            )
            self.assertFalse(
                invalid_simplifications,
                "Invalid simplifications {} for locale {}".format(
                    ", ".join(map(repr, invalid_simplifications)), self.shortname
                )
                + ", each simplification must be a single string to string mapping",
            )

    def then_relative_type_is_valid(self):
        if "relative-type" in self.info:
            relative_type_dict = self.info["relative-type"]
            self.assertIsInstance(
                relative_type_dict,
                dict,
                "Invalid type for relative-type: {} for locale {}".format(
                    type(relative_type_dict).__name__, self.shortname
                ),
            )
            invalid_relative_type = list(
                filter(is_invalid_relative_mapping, relative_type_dict.items())
            )
            self.assertFalse(
                invalid_relative_type,
                "Invalid relative-type mappings {} for locale {}".format(
                    ", ".join(map(repr, invalid_relative_type)), self.shortname
                )
                + ", each mapping must be a string to list (of strings) mapping",
            )

    def then_relative_type_regex_is_valid(self):
        if "relative-type-regex" in self.info:
            relative_type_dict = self.info["relative-type-regex"]
            self.assertIsInstance(
                relative_type_dict,
                dict,
                "Invalid type for relative-type-regex: {} for locale {}".format(
                    type(relative_type_dict).__name__, self.shortname
                ),
            )
            invalid_relative_type_regex = list(
                filter(is_invalid_relative_regex_mapping, relative_type_dict.items())
            )
            self.assertFalse(
                invalid_relative_type_regex,
                "Invalid relative-type-regex mappings {} for locale {}".format(
                    ", ".join(map(repr, invalid_relative_type_regex)), self.shortname
                )
                + ", each mapping must be a string to list (of strings) mapping",
            )

    def then_no_word_spacing_is_valid(self):
        if "no_word_spacing" in self.info:
            no_word_spacing = self.info["no_word_spacing"]
            self.assertIsInstance(
                no_word_spacing,
                str,
                "Invalid type for no_word_spacing: {} for locale {}".format(
                    type(no_word_spacing).__name__, self.shortname
                ),
            )
            self.assertIn(
                no_word_spacing,
                ["True", "False"],
                "Invalid no_word_spacing {} for {}".format(
                    no_word_spacing, self.shortname
                ),
            )



info = {
    "name": "az-Cyrl",
    "date_order": "DMY",
    "january": [
        "јан",
        "јанвар"
    ],
    "february": [
        "фев",
        "феврал"
    ],
    "march": [
        "мар",
        "март"
    ],
    "april": [
        "апр",
        "апрел"
    ],
    "may": [
        "май"
    ],
    "june": [
        "ијн",
        "ијун"
    ],
    "july": [
        "ијл",
        "ијул"
    ],
    "august": [
        "авг",
        "август"
    ],
    "september": [
        "сен",
        "сентјабр"
    ],
    "october": [
        "окт",
        "октјабр"
    ],
    "november": [
        "ној",
        "нојабр"
    ],
    "december": [
        "дек",
        "декабр"
    ],
    "monday": [
        "базар ертәси",
        "бе"
    ],
    "tuesday": [
        "ча",
        "чәршәнбә ахшамы"
    ],
    "wednesday": [
        "ч",
        "чәршәнбә"
    ],
    "thursday": [
        "ҹа",
        "ҹүмә ахшамы"
    ],
    "friday": [
        "ҹ",
        "ҹүмә"
    ],
    "saturday": [
        "ш",
        "шәнбә"
    ],
    "sunday": [
        "б",
        "базар"
    ],
    "am": [
        "ам"
    ],
    "pm": [
        "пм"
    ],
    "year": [
        "year"
    ],
    "month": [
        "month"
    ],
    "week": [
        "week"
    ],
    "day": [
        "day"
    ],
    "hour": [
        "hour"
    ],
    "minute": [
        "minute"
    ],
    "second": [
        "second"
    ],
    "relative-type": {
        "0 day ago": [
            "today"
        ],
        "0 hour ago": [
            "this hour"
        ],
        "0 minute ago": [
            "this minute"
        ],
        "0 month ago": [
            "this month"
        ],
        "0 second ago": [
            "now"
        ],
        "0 week ago": [
            "this week"
        ],
        "0 year ago": [
            "this year"
        ],
        "1 day ago": [
            "yesterday"
        ],
        "1 month ago": [
            "last month"
        ],
        "1 week ago": [
            "last week"
        ],
        "1 year ago": [
            "last year"
        ],
        "in 1 day": [
            "tomorrow"
        ],
        "in 1 month": [
            "next month"
        ],
        "in 1 week": [
            "next week"
        ],
        "in 1 year": [
            "next year"
        ]
    },
    "locale_specific": {},
    "skip": [
        " ",
        "'",
        ",",
        "-",
        ".",
        "/",
        ";",
        "@",
        "[",
        "]",
        "|",
        "，"
    ]
}



from datetime import datetime, tzinfo

import pytest
from parameterized import param, parameterized

from dateparser import DateDataParser, parse
from dateparser.conf import SettingValidationError, apply_settings, settings
from tests import BaseTestCase


def test_function(settings=None):
    return settings


class TimeZoneSettingsTest(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.given_ds = NotImplemented
        self.result = NotImplemented
        self.timezone = NotImplemented
        self.confs = NotImplemented

    @parameterized.expand(
        [
            param(
                "12 Feb 2015 10:30 PM +0100",
                datetime(2015, 2, 12, 22, 30),
                r"UTC\+01:00",
            ),
            param("12 Feb 2015 4:30 PM EST", datetime(2015, 2, 12, 16, 30), "EST"),
            param("12 Feb 2015 8:30 PM PKT", datetime(2015, 2, 12, 20, 30), "PKT"),
            param("12 Feb 2015 8:30 PM ACT", datetime(2015, 2, 12, 20, 30), "ACT"),
        ]
    )
    def test_should_return_and_assert_tz(self, ds, dt, tz):
        self.given(ds)
        self.given_configurations({})
        self.when_date_is_parsed()
        self.then_date_is_tz_aware()
        self.then_date_is(dt)
        self.then_timezone_is(tz)

    @parameterized.expand(
        [
            param("12 Feb 2015 4:30 PM EST", datetime(2015, 2, 12, 16, 30), "EST"),
            param("12 Feb 2015 8:30 PM PKT", datetime(2015, 2, 12, 20, 30), "PKT"),
            param("12 Feb 2015 8:30 PM ACT", datetime(2015, 2, 12, 20, 30), "ACT"),
            param("12 Feb 2015 8:30 PM", datetime(2015, 2, 12, 20, 30), ""),
        ]
    )
    def test_only_return_explicit_timezone(self, ds, dt, tz):
        self.given(ds)
        self.given_configurations({})
        self.when_date_is_parsed()
        self.then_date_is(dt)
        if tz:
            self.then_date_is_tz_aware()
            self.then_timezone_is(tz)
        else:
            self.then_date_is_not_tz_aware()

    @parameterized.expand(
        [
            param(
                "12 Feb 2015 4:30 PM EST",
                datetime(2015, 2, 12, 16, 30),
            ),
            param(
                "12 Feb 2015 8:30 PM PKT",
                datetime(2015, 2, 12, 20, 30),
            ),
            param(
                "12 Feb 2015 8:30 PM ACT",
                datetime(2015, 2, 12, 20, 30),
            ),
            param(
                "12 Feb 2015 8:30 PM +0100",
                datetime(2015, 2, 12, 20, 30),
            ),
        ]
    )
    def test_should_return_naive_if_RETURN_AS_TIMEZONE_AWARE_IS_FALSE(self, ds, dt):
        self.given(ds)
        self.given_configurations({"RETURN_AS_TIMEZONE_AWARE": False})
        self.when_date_is_parsed()
        self.then_date_is(dt)
        self.then_date_is_not_tz_aware()

    def then_timezone_is(self, tzname):
        self.assertEqual(self.result.tzinfo.tzname(""), tzname)

    def given(self, ds):
        self.given_ds = ds

    def given_configurations(self, confs):
        if "TIMEZONE" not in confs:
            confs.update({"TIMEZONE": "local"})

        self.confs = settings.replace(**confs)

    def when_date_is_parsed(self):
        self.result = parse(self.given_ds, settings=(self.confs or {}))

    def then_date_is_tz_aware(self):
        self.assertIsInstance(self.result.tzinfo, tzinfo)

    def then_date_is_not_tz_aware(self):
        self.assertIsNone(self.result.tzinfo)

    def then_date_is(self, date):
        dtc = self.result.replace(tzinfo=None)
        self.assertEqual(dtc, date)


class SettingsTest(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.default_settings = settings

    def test_apply_settings_should_return_default_settings_when_no_settings_are_supplied_to_the_decorated_function(
        self,
    ):
        test_func = apply_settings(test_function)
        self.assertEqual(test_func(), self.default_settings)

    def test_apply_settings_should_return_non_default_settings_when_settings_are_supplied_to_the_decorated_function(
        self,
    ):
        test_func = apply_settings(test_function)
        self.assertNotEqual(
            test_func(settings={"PREFER_DATES_FROM": "past"}), self.default_settings
        )

    def test_apply_settings_shouldnt_create_new_settings_when_same_settings_are_supplied_to_the_decorated_function_more_than_once(  # noqa E501
        self,
    ):
        test_func = apply_settings(test_function)
        settings_once = test_func(settings={"PREFER_DATES_FROM": "past"})
        settings_twice = test_func(settings={"PREFER_DATES_FROM": "past"})
        self.assertEqual(settings_once, settings_twice)

    def test_apply_settings_should_return_default_settings_when_called_with_no_settings_after_once_called_with_settings_supplied_to_the_decorated_function(  # noqa E501
        self,
    ):
        test_func = apply_settings(test_function)
        settings_once = test_func(settings={"PREFER_DATES_FROM": "past"})
        settings_twice = test_func()
        self.assertNotEqual(settings_once, self.default_settings)
        self.assertEqual(settings_twice, self.default_settings)


class InvalidSettingsTest(BaseTestCase):
    def setUp(self):
        super().setUp()

    def test_error_is_raised_when_none_is_passed_in_settings(self):
        test_func = apply_settings(test_function)
        with self.assertRaisesRegex(TypeError, r"Invalid.*None\}"):
            test_func(settings={"PREFER_DATES_FROM": None})

        with self.assertRaisesRegex(TypeError, r"Invalid.*None\}"):
            test_func(settings={"TIMEZONE": None})

        with self.assertRaisesRegex(TypeError, r"Invalid.*None\}"):
            test_func(settings={"TO_TIMEZONE": None})

    def test_error_is_raised_for_invalid_type_settings(self):
        test_func = apply_settings(test_function)
        try:
            test_func(settings=["current_period", False, "current"])
        except Exception as error:
            self.error = error
            self.then_error_was_raised(
                TypeError,
                ["settings can only be either dict or instance of Settings class"],
            )

    def test_check_settings_wrong_setting_name(self):
        with self.assertRaisesRegex(
            SettingValidationError, r".* is not a valid setting"
        ):
            DateDataParser(settings={"AAAAA": "foo"})

    @parameterized.expand(
        [
            param("DATE_ORDER", 2, "YYY", "MDY"),
            param(
                "TIMEZONE", False, "", "Europe/Madrid"
            ),  # should we check valid timezones?
            param(
                "TO_TIMEZONE", True, "", "Europe/Madrid"
            ),  # should we check valid timezones?
            param("RETURN_AS_TIMEZONE_AWARE", "false", "", True),
            param("PREFER_DAY_OF_MONTH", False, "current_period", "current"),
            param("PREFER_DATES_FROM", True, "current", "current_period"),
            param("RELATIVE_BASE", "yesterday", "", datetime.now()),
            param("SKIP_TOKENS", "foo", "", ["foo"]),
            param("REQUIRE_PARTS", "day", "", ["month", "day"]),
            param("PARSERS", "absolute-time", "", ["absolute-time", "no-spaces-time"]),
            param("STRICT_PARSING", "true", "", True),
            param("RETURN_TIME_AS_PERIOD", "false", "", True),
            param("PREFER_LOCALE_DATE_ORDER", "true", "", False),
            param("NORMALIZE", "true", "", True),
            param("FUZZY", "true", "", False),
            param("PREFER_LOCALE_DATE_ORDER", "false", "", True),
            param("DEFAULT_LANGUAGES", "en", "", ["en"]),
            param("LANGUAGE_DETECTION_CONFIDENCE_THRESHOLD", "1", "", 0.5),
        ]
    )
    def test_check_settings(self, setting, wrong_type, wrong_value, valid_value):
        with self.assertRaisesRegex(
            SettingValidationError,
            r'"{}" must be .*, not "{}".'.format(setting, type(wrong_type).__name__),
        ):
            DateDataParser(settings={setting: wrong_type})

        if wrong_value:
            with self.assertRaisesRegex(
                SettingValidationError,
                r'"{}" is not a valid value for "{}", it should be: .*'.format(
                    str(wrong_value).replace("[", "\\[").replace("]", "\\]"), setting
                ),
            ):
                DateDataParser(settings={setting: wrong_value})

        # check that a valid value doesn't raise an error
        assert DateDataParser(settings={setting: valid_value})

    def test_check_settings_extra_check_require_parts(self):
        with self.assertRaisesRegex(
            SettingValidationError,
            r'"REQUIRE_PARTS" setting contains invalid values: time',
        ):
            DateDataParser(settings={"REQUIRE_PARTS": ["time", "day"]})
        with self.assertRaisesRegex(
            SettingValidationError,
            r'There are repeated values in the "REQUIRE_PARTS" setting',
        ):
            DateDataParser(settings={"REQUIRE_PARTS": ["month", "day", "month"]})

    def test_check_settings_extra_check_parsers(self):
        with self.assertRaisesRegex(
            SettingValidationError,
            r'Found unknown parsers in the "PARSERS" setting: no-spaces',
        ):
            DateDataParser(settings={"PARSERS": ["absolute-time", "no-spaces"]})

        with self.assertRaisesRegex(
            SettingValidationError,
            r'There are repeated values in the "PARSERS" setting',
        ):
            DateDataParser(
                settings={"PARSERS": ["absolute-time", "timestamp", "absolute-time"]}
            )

    def test_check_settings_extra_check_confidence_threshold(self):
        with self.assertRaisesRegex(
            SettingValidationError,
            r"1.1 is not a valid value for "
            r"LANGUAGE_DETECTION_CONFIDENCE_THRESHOLD. It can take values "
            r"between 0 and 1",
        ):
            DateDataParser(settings={"LANGUAGE_DETECTION_CONFIDENCE_THRESHOLD": 1.1})

    def test_check_settings_extra_check_default_languages(self):
        with self.assertRaisesRegex(
            SettingValidationError,
            "Found invalid languages in the 'DEFAULT_LANGUAGES' setting: 'abcd'",
        ):
            DateDataParser(settings={"DEFAULT_LANGUAGES": ["abcd"]})


@pytest.mark.parametrize(
    "date_string,expected_result",
    [
        # Note that these results are "valid" but probably they shouldn't be considered
        ("2020", datetime(1900, 1, 1, 20, 2)),
        ("333", datetime(1900, 1, 1, 3, 3, 3)),
        ("1000", datetime(1900, 1, 1, 10, 0)),
        ("100000", datetime(1900, 1, 1, 10, 0)),
    ],
)
def test_no_spaces_strict_parsing(date_string, expected_result):
    parser = DateDataParser(
        settings={"PARSERS": ["no-spaces-time"], "STRICT_PARSING": False}
    )
    assert parser.get_date_data(date_string)["date_obj"] == expected_result

    parser = DateDataParser(
        settings={"PARSERS": ["no-spaces-time"], "STRICT_PARSING": True}
    )
    assert parser.get_date_data(date_string)["date_obj"] is None


def detect_languages(text, confidence_threshold):
    if confidence_threshold > 0.5:
        return ["en"]
    else:
        return ["fr"]


def test_confidence_threshold_setting_is_applied():
    ddp = DateDataParser(
        detect_languages_function=detect_languages,
        settings={"LANGUAGE_DETECTION_CONFIDENCE_THRESHOLD": 0.6},
    )
    assert ddp.get_date_data("21/06/2020").locale == "en"

    ddp2 = DateDataParser(
        detect_languages_function=detect_languages,
        settings={"LANGUAGE_DETECTION_CONFIDENCE_THRESHOLD": 0.4},
    )
    assert ddp2.get_date_data("21/06/2020").locale == "fr"



import pickle
from datetime import datetime

from dateparser import parse
from dateparser.timezone_parser import StaticTzInfo
from tests import BaseTestCase


class TestPickle(BaseTestCase):
    def test_pickling_basic_parsed_object(self):
        now = parse("now")
        self.assertIsInstance(now, datetime)
        pickle_dumps = pickle.dumps(now)
        self.assertIsInstance(pickle_dumps, bytes)

    def test_pickling_parsed_object_with_tz(self):
        now_in_utc = parse("now in UTC")
        self.assertIsInstance(now_in_utc, datetime)
        self.assertIsInstance(now_in_utc.tzinfo, StaticTzInfo)
        pickle_dumps = pickle.dumps(now_in_utc)
        self.assertIsInstance(pickle_dumps, bytes)

    def test_unpickling_basic_parsed_object(self):
        now = parse("now")
        pickle_dumps = pickle.dumps(now)
        pickle_loads = pickle.loads(pickle_dumps)
        self.assertIsInstance(pickle_loads, datetime)

    def test_unpickling_parsed_object_with_tz(self):
        now_in_utc = parse("now in UTC")
        pickle_dumps = pickle.dumps(now_in_utc)
        pickle_loads = pickle.loads(pickle_dumps)
        self.assertIsInstance(pickle_loads, datetime)
        self.assertIsInstance(pickle_loads.tzinfo, StaticTzInfo)



import contextlib
import datetime
import io
import tempfile
from typing import List, TypeVar

import atheris

T = TypeVar("T")


class EnhancedFuzzedDataProvider(atheris.FuzzedDataProvider):
    def ConsumeRandomBytes(self) -> bytes:
        return self.ConsumeBytes(self.ConsumeIntInRange(0, self.remaining_bytes()))

    def ConsumeRandomString(self) -> str:
        return self.ConsumeUnicodeNoSurrogates(
            self.ConsumeIntInRange(0, self.remaining_bytes())
        )

    def ConsumeRemainingString(self) -> str:
        return self.ConsumeUnicodeNoSurrogates(self.remaining_bytes())

    def ConsumeRemainingBytes(self) -> bytes:
        return self.ConsumeBytes(self.remaining_bytes())

    def ConsumeSublist(self, source: List[T]) -> List[T]:
        """
        Returns a shuffled sub-list of the given list of len [1, len(source)]
        """
        chosen = [elem for elem in source if self.ConsumeBool()]

        # Shuffle
        for i in range(len(chosen) - 1, 1, -1):
            j = self.ConsumeIntInRange(0, i)
            chosen[i], chosen[j] = chosen[j], chosen[i]

        return chosen or [self.PickValueInList(source)]

    def ConsumeDate(self) -> datetime.datetime:
        try:
            return datetime.datetime.fromtimestamp(self.ConsumeFloat())
        except (OverflowError, OSError, ValueError):
            return datetime.datetime(year=1970, month=1, day=1)

    @contextlib.contextmanager
    def ConsumeMemoryFile(
        self, all_data: bool = False, as_bytes: bool = True
    ) -> io.BytesIO:
        if all_data:
            file_data = (
                self.ConsumeRemainingBytes()
                if as_bytes
                else self.ConsumeRemainingString()
            )
        else:
            file_data = (
                self.ConsumeRandomBytes() if as_bytes else self.ConsumeRandomString()
            )

        file = io.BytesIO(file_data) if as_bytes else io.StringIO(file_data)
        yield file
        file.close()

    @contextlib.contextmanager
    def ConsumeTemporaryFile(
        self, suffix: str, all_data: bool = False, as_bytes: bool = True
    ) -> str:
        if all_data:
            file_data = (
                self.ConsumeRemainingBytes()
                if as_bytes
                else self.ConsumeRemainingString()
            )
        else:
            file_data = (
                self.ConsumeRandomBytes() if as_bytes else self.ConsumeRandomString()
            )

        mode = "w+b" if as_bytes else "w+"
        tfile = tempfile.NamedTemporaryFile(mode=mode, suffix=suffix)
        tfile.write(file_data)
        tfile.seek(0)
        tfile.flush()
        yield tfile.name
        tfile.close()



#!/usr/bin/env python

import os
import unittest
from collections import OrderedDict
from datetime import datetime, timedelta
from datetime import timezone as dttz
from itertools import product
from time import tzset
from unittest.mock import Mock, patch

from parameterized import param, parameterized

import dateparser
from dateparser import date
from dateparser.conf import Settings
from dateparser.date import DateData
from tests import BaseTestCase


class TestDateRangeFunction(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.result = NotImplemented

    @parameterized.expand(
        [
            param(
                begin=datetime(2014, 6, 15),
                end=datetime(2014, 6, 25),
                expected_length=10,
            )
        ]
    )
    def test_date_range(self, begin, end, expected_length):
        self.when_date_range_generated(begin, end)
        self.then_range_length_is(expected_length)
        self.then_all_dates_in_range_are_present(begin, end)
        self.then_range_is_in_ascending_order()

    @parameterized.expand(
        [
            param(
                begin=datetime(2014, 4, 15),
                end=datetime(2014, 6, 25),
                expected_months=[(2014, 4), (2014, 5), (2014, 6)],
            ),
            param(
                begin=datetime(2014, 4, 25),
                end=datetime(2014, 5, 5),
                expected_months=[(2014, 4), (2014, 5)],
            ),
            param(
                begin=datetime(2014, 4, 5),
                end=datetime(2014, 4, 25),
                expected_months=[(2014, 4)],
            ),
            param(
                begin=datetime(2014, 4, 25),
                end=datetime(2014, 6, 5),
                expected_months=[(2014, 4), (2014, 5), (2014, 6)],
            ),
        ]
    )
    def test_one_date_for_each_month(self, begin, end, expected_months):
        self.when_date_range_generated(begin, end, months=1)
        self.then_expected_months_are(expected_months)

    @parameterized.expand(
        [
            "year",
            "month",
            "week",
            "day",
            "hour",
            "minute",
            "second",
        ]
    )
    def test_should_reject_easily_mistaken_dateutil_arguments(self, invalid_period):
        self.when_date_range_generated(
            begin=datetime(2014, 6, 15),
            end=datetime(2014, 6, 25),
            **{invalid_period: 1}
        )
        self.then_period_was_rejected(invalid_period)

    def when_date_range_generated(self, begin, end, **size):
        try:
            self.result = list(date.date_range(begin, end, **size))
        except Exception as error:
            self.error = error

    def then_expected_months_are(self, expected):
        self.assertEqual(expected, [(d.year, d.month) for d in self.result])

    def then_range_length_is(self, expected_length):
        self.assertEqual(expected_length, len(self.result))

    def then_all_dates_in_range_are_present(self, begin, end):
        date_under_test = begin
        while date_under_test < end:
            self.assertIn(date_under_test, self.result)
            date_under_test += timedelta(days=1)

    def then_range_is_in_ascending_order(self):
        for i in range(len(self.result) - 1):
            self.assertLess(self.result[i], self.result[i + 1])

    def then_period_was_rejected(self, period):
        self.then_error_was_raised(ValueError, ["Invalid argument: {}".format(period)])


class TestGetIntersectingPeriodsFunction(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.result = NotImplemented

    @parameterized.expand(
        [param(low=datetime(2014, 6, 15), high=datetime(2014, 6, 16), length=1)]
    )
    def test_date_arguments_and_date_range_with_default_post_days(
        self, low, high, length
    ):
        self.when_intersecting_period_calculated(low, high, period_size="day")
        self.then_all_dates_in_range_are_present(begin=low, end=high)
        self.then_date_range_length_is(length)

    @parameterized.expand(
        [
            param(
                low=datetime(2014, 4, 15),
                high=datetime(2014, 6, 25),
                expected_results=[
                    datetime(2014, 4, 1),
                    datetime(2014, 5, 1),
                    datetime(2014, 6, 1),
                ],
            ),
            param(
                low=datetime(2014, 4, 25),
                high=datetime(2014, 5, 5),
                expected_results=[datetime(2014, 4, 1), datetime(2014, 5, 1)],
            ),
            param(
                low=datetime(2014, 4, 5),
                high=datetime(2014, 4, 25),
                expected_results=[datetime(2014, 4, 1)],
            ),
            param(
                low=datetime(2014, 4, 25),
                high=datetime(2014, 6, 5),
                expected_results=[
                    datetime(2014, 4, 1),
                    datetime(2014, 5, 1),
                    datetime(2014, 6, 1),
                ],
            ),
            param(
                low=datetime(2014, 4, 25),
                high=datetime(2014, 4, 25),
                expected_results=[],
            ),
            param(
                low=datetime(2014, 12, 31),
                high=datetime(2015, 1, 1),
                expected_results=[datetime(2014, 12, 1)],
            ),
        ]
    )
    def test_dates_in_intersecting_period_should_use_first_day_when_period_is_month(
        self, low, high, expected_results
    ):
        self.when_intersecting_period_calculated(low, high, period_size="month")
        self.then_results_are(expected_results)

    @parameterized.expand(
        [
            param(
                low=datetime(2012, 4, 18),
                high=datetime(2014, 9, 22),
                expected_results=[
                    datetime(2012, 1, 1, 0, 0),
                    datetime(2013, 1, 1, 0, 0),
                    datetime(2014, 1, 1, 0, 0),
                ],
            ),
            param(
                low=datetime(2013, 8, 5),
                high=datetime(2014, 5, 15),
                expected_results=[
                    datetime(2013, 1, 1, 0, 0),
                    datetime(2014, 1, 1, 0, 0),
                ],
            ),
            param(
                low=datetime(2008, 4, 5),
                high=datetime(2010, 1, 1),
                expected_results=[
                    datetime(2008, 1, 1, 0, 0),
                    datetime(2009, 1, 1, 0, 0),
                ],
            ),
            param(
                low=datetime(2014, 1, 1),
                high=datetime(2016, 8, 22),
                expected_results=[
                    datetime(2014, 1, 1, 0, 0),
                    datetime(2015, 1, 1, 0, 0),
                    datetime(2016, 1, 1, 0, 0),
                ],
            ),
            param(
                low=datetime(2001, 7, 11),
                high=datetime(2001, 10, 16),
                expected_results=[datetime(2001, 1, 1, 0, 0)],
            ),
            param(
                low=datetime(2017, 1, 1), high=datetime(2017, 1, 1), expected_results=[]
            ),
        ]
    )
    def test_dates_in_intersecting_period_should_use_first_month_and_first_day_when_period_is_year(
        self, low, high, expected_results
    ):
        self.when_intersecting_period_calculated(low, high, period_size="year")
        self.then_results_are(expected_results)

    @parameterized.expand(
        [
            param(
                low=datetime(2014, 4, 15),
                high=datetime(2014, 5, 15),
                period_size="month",
                expected_results=[datetime(2014, 4, 1), datetime(2014, 5, 1)],
            ),
            param(
                low=datetime(2014, 10, 30, 4, 30),
                high=datetime(2014, 11, 7, 5, 20),
                period_size="week",
                expected_results=[datetime(2014, 10, 27), datetime(2014, 11, 3)],
            ),
            param(
                low=datetime(2014, 8, 13, 13, 21),
                high=datetime(2014, 8, 14, 14, 7),
                period_size="day",
                expected_results=[datetime(2014, 8, 13), datetime(2014, 8, 14)],
            ),
            param(
                low=datetime(2014, 5, 11, 22, 4),
                high=datetime(2014, 5, 12, 0, 5),
                period_size="hour",
                expected_results=[
                    datetime(2014, 5, 11, 22, 0),
                    datetime(2014, 5, 11, 23, 0),
                    datetime(2014, 5, 12, 0, 0),
                ],
            ),
            param(
                low=datetime(2014, 4, 25, 11, 11, 11),
                high=datetime(2014, 4, 25, 11, 12, 11),
                period_size="minute",
                expected_results=[
                    datetime(2014, 4, 25, 11, 11, 0),
                    datetime(2014, 4, 25, 11, 12, 0),
                ],
            ),
            param(
                low=datetime(2014, 12, 31, 23, 59, 58, 500),
                high=datetime(2014, 12, 31, 23, 59, 59, 600),
                period_size="second",
                expected_results=[
                    datetime(2014, 12, 31, 23, 59, 58, 0),
                    datetime(2014, 12, 31, 23, 59, 59, 0),
                ],
            ),
        ]
    )
    def test_periods(self, low, high, period_size, expected_results):
        self.when_intersecting_period_calculated(low, high, period_size=period_size)
        self.then_results_are(expected_results)

    @parameterized.expand(
        [
            param("years"),
            param("months"),
            param("days"),
            param("hours"),
            param("minutes"),
            param("seconds"),
            param("microseconds"),
            param("some_period"),
        ]
    )
    def test_should_reject_easily_mistaken_dateutil_arguments(self, period_size):
        self.when_intersecting_period_calculated(
            low=datetime(2014, 6, 15),
            high=datetime(2014, 6, 25),
            period_size=period_size,
        )
        self.then_error_was_raised(ValueError, ["Invalid period: " + str(period_size)])

    @parameterized.expand(
        [
            param(
                low=datetime(2014, 4, 15),
                high=datetime(2014, 4, 14),
                period_size="month",
            ),
            param(
                low=datetime(2014, 4, 25),
                high=datetime(2014, 4, 25),
                period_size="month",
            ),
        ]
    )
    def test_empty_period(self, low, high, period_size):
        self.when_intersecting_period_calculated(low, high, period_size)
        self.then_period_is_empty()

    def when_intersecting_period_calculated(self, low, high, period_size):
        try:
            self.result = list(
                date.get_intersecting_periods(low, high, period=period_size)
            )
        except Exception as error:
            self.error = error

    def then_results_are(self, expected_results):
        self.assertEqual(expected_results, self.result)

    def then_date_range_length_is(self, size):
        self.assertEqual(size, len(self.result))

    def then_all_dates_in_range_are_present(self, begin, end):
        date_under_test = begin
        while date_under_test < end:
            self.assertIn(date_under_test, self.result)
            date_under_test += timedelta(days=1)

    def then_period_is_empty(self):
        self.assertEqual([], self.result)


class TestParseWithFormatsFunction(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.result = NotImplemented

    @parameterized.expand(
        [
            param(date_string="yesterday", date_formats=["%Y-%m-%d"]),
        ]
    )
    def test_date_with_not_matching_format_is_not_parsed(
        self, date_string, date_formats
    ):
        self.when_date_is_parsed_with_formats(date_string, date_formats)
        self.then_date_was_not_parsed()

    @parameterized.expand(
        [
            param(
                date_string="25-03-14",
                date_formats=["%d-%m-%y"],
                expected_result=datetime(2014, 3, 25),
            ),
        ]
    )
    def test_should_parse_date(self, date_string, date_formats, expected_result):
        self.when_date_is_parsed_with_formats(date_string, date_formats)
        self.then_date_was_parsed()
        self.then_parsed_period_is("day")
        self.then_parsed_date_is(expected_result)

    @parameterized.expand(
        [
            param(
                date_string="09.16",
                date_formats=["%m.%d"],
                expected_month=9,
                expected_day=16,
            ),
        ]
    )
    def test_should_use_current_year_for_dates_without_year(
        self, date_string, date_formats, expected_month, expected_day
    ):
        self.given_now(2015, 2, 4)
        self.when_date_is_parsed_with_formats(date_string, date_formats)
        self.then_date_was_parsed()
        self.then_parsed_period_is("day")
        self.then_parsed_date_is(datetime(2015, expected_month, expected_day))

    @parameterized.expand(
        [
            param(
                date_string="August 2014",
                date_formats=["%B %Y"],
                expected_year=2014,
                expected_month=8,
                today_day=12,
                prefer_day_of_month="first",
                expected_day=1,
            ),
            param(
                date_string="August 2014",
                date_formats=["%B %Y"],
                expected_year=2014,
                expected_month=8,
                today_day=12,
                prefer_day_of_month="last",
                expected_day=31,
            ),
            param(
                date_string="August 2014",
                date_formats=["%B %Y"],
                expected_year=2014,
                expected_month=8,
                today_day=12,
                prefer_day_of_month="current",
                expected_day=12,
            ),
        ]
    )
    def test_should_use_correct_day_from_settings_for_dates_without_day(
        self,
        date_string,
        date_formats,
        expected_year,
        expected_month,
        today_day,
        prefer_day_of_month,
        expected_day,
    ):
        self.given_now(2014, 8, today_day)
        settings_mod = Settings()
        settings_mod.PREFER_DAY_OF_MONTH = prefer_day_of_month
        self.when_date_is_parsed_with_formats(date_string, date_formats, settings_mod)
        self.then_date_was_parsed()
        self.then_parsed_period_is("month")
        self.then_parsed_date_is(
            datetime(year=expected_year, month=expected_month, day=expected_day)
        )

    @parameterized.expand(
        [
            param(
                date_string="2014",
                date_formats=["%Y"],
                expected_year=2014,
                prefer_month_of_year="first",
                current_month=7,
                expected_month=1,
                expected_day=1,
            ),
            param(
                date_string="2014",
                date_formats=["%Y"],
                expected_year=2014,
                prefer_month_of_year="current",
                current_month=7,
                expected_month=7,
                expected_day=1,
            ),
            param(
                date_string="2014",
                date_formats=["%Y"],
                expected_year=2014,
                prefer_month_of_year="last",
                current_month=7,
                expected_month=12,
                expected_day=1,
            ),
        ]
    )
    def test_should_use_correct_month_from_settings_for_dates_without_month(
        self,
        date_string,
        date_formats,
        expected_year,
        prefer_month_of_year,
        current_month,
        expected_month,
        expected_day,
    ):
        self.given_now(2014, current_month, 1)
        settings_mod = Settings()
        settings_mod.PREFER_MONTH_OF_YEAR = prefer_month_of_year
        self.when_date_is_parsed_with_formats(date_string, date_formats, settings_mod)
        self.then_date_was_parsed()
        self.then_parsed_period_is("year")
        self.then_parsed_date_is(
            datetime(year=expected_year, month=expected_month, day=expected_day)
        )

    @parameterized.expand(
        [
            param(
                date_string="2014",
                date_formats=["%Y"],
                current_day=15,
                current_month=4,
                prefer_day_of_month="last",
                prefer_month_of_year="last",
                expected_year=2014,
                expected_month=12,
                expected_day=31,
            )
        ]
    )
    def test_should_use_correct_day_n_month_from_settings_for_dates_without_day_n_month(
        self,
        date_string,
        date_formats,
        current_day,
        current_month,
        prefer_day_of_month,
        prefer_month_of_year,
        expected_year,
        expected_month,
        expected_day,
    ):
        self.given_now(2014, current_month, current_day)
        settings_mod = Settings()
        settings_mod.PREFER_DAY_OF_MONTH = prefer_day_of_month
        settings_mod.PREFER_MONTH_OF_YEAR = prefer_month_of_year
        self.when_date_is_parsed_with_formats(date_string, date_formats, settings_mod)
        self.then_date_was_parsed()
        self.then_parsed_period_is("year")
        self.then_parsed_date_is(
            datetime(year=expected_year, month=expected_month, day=expected_day)
        )

    def given_now(self, year, month, day, **time):
        now = datetime(year, month, day, **time)
        datetime_mock = Mock(wraps=datetime)
        datetime_mock.utcnow = Mock(return_value=now)
        datetime_mock.now = Mock(return_value=now)
        datetime_mock.today = Mock(return_value=now)
        self.add_patch(patch("dateparser.date.datetime", new=datetime_mock))
        self.add_patch(patch("dateparser.utils.datetime", new=datetime_mock))

    def when_date_is_parsed_with_formats(
        self, date_string, date_formats, custom_settings=None
    ):
        self.result = date.parse_with_formats(
            date_string, date_formats, custom_settings or Settings()
        )

    def then_date_was_not_parsed(self):
        self.assertIsNotNone(self.result)
        self.assertIsNone(self.result["date_obj"])

    def then_date_was_parsed(self):
        self.assertIsNotNone(self.result)
        self.assertIsNotNone(self.result["date_obj"])

    def then_parsed_date_is(self, date_obj):
        self.assertEqual(date_obj.date(), self.result["date_obj"].date())

    def then_parsed_period_is(self, period):
        self.assertEqual(period, self.result["period"])


class TestDateDataParser(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.parser = NotImplemented
        self.result = NotImplemented
        self.multiple_results = NotImplemented

    @parameterized.expand(
        [
            param("10:04am EDT"),
        ]
    )
    def test_time_without_date_should_use_today(self, date_string):
        self.given_parser(settings={"RELATIVE_BASE": datetime(2020, 7, 19)})
        self.when_date_string_is_parsed(date_string)
        self.then_date_was_parsed()
        self.then_parsed_date_is(datetime(2020, 7, 19).date())

    @parameterized.expand(
        [
            # Today
            param("today", days_ago=0),
            param("Today", days_ago=0),
            param("TODAY", days_ago=0),
            param("Сегодня", days_ago=0),
            param("Hoje", days_ago=0),
            param("Oggi", days_ago=0),
            # Yesterday
            param("yesterday", days_ago=1),
            param(" Yesterday \n", days_ago=1),
            param("Ontem", days_ago=1),
            param("Ieri", days_ago=1),
            param("вчера", days_ago=1),
            param("снощи", days_ago=1),
            # Day before yesterday
            param("the day before yesterday", days_ago=2),
            param("The DAY before Yesterday", days_ago=2),
            param("Anteontem", days_ago=2),
            param("Avant-hier", days_ago=2),
            param("вчера", days_ago=1),
            param("снощи", days_ago=1),
        ]
    )
    def test_temporal_nouns_are_parsed(self, date_string, days_ago):
        self.given_parser()
        self.when_date_string_is_parsed(date_string)
        self.then_date_was_parsed()
        self.then_date_is_n_days_ago(days=days_ago)

    def test_should_not_assume_language_too_early(self):
        dates_to_parse = OrderedDict(
            [
                ("07/07/2014", datetime(2014, 7, 7).date()),  # any language
                ("07.jul.2014 | 12:52", datetime(2014, 7, 7).date()),  # en, es, pt, nl
                ("07.ago.2014 | 12:52", datetime(2014, 8, 7).date()),  # es, it, pt
                (
                    "07.feb.2014 | 12:52",
                    datetime(2014, 2, 7).date(),
                ),  # en, de, es, it, nl, ro
                ("07.ene.2014 | 12:52", datetime(2014, 1, 7).date()),
            ]
        )  # es

        self.given_parser(
            restrict_to_languages=["en", "de", "fr", "it", "pt", "nl", "ro", "es", "ru"]
        )
        self.when_multiple_dates_are_parsed(dates_to_parse.keys())
        self.then_all_results_were_parsed()
        self.then_parsed_dates_are(list(dates_to_parse.values()))

    @parameterized.expand(
        [
            param(date_string="11 Marzo, 2014", locale="es"),
            param(date_string="13 Septiembre, 2014", locale="es"),
            param(date_string="Сегодня", locale="ru"),
            param(date_string="Avant-hier", locale="fr"),
            param(date_string="Anteontem", locale="pt"),
            param(date_string="ธันวาคม 11, 2014, 08:55:08 PM", locale="th"),
            param(date_string="Anteontem", locale="pt"),
            param(date_string="14 aprilie 2014", locale="ro"),
            param(date_string="11 Ağustos, 2014", locale="tr"),
            param(date_string="pon 16. čer 2014 10:07:43", locale="cs"),
            param(date_string="24 януари 2015г.", locale="bg"),
        ]
    )
    def test_returned_detected_locale_should_be(self, date_string, locale):
        self.given_parser()
        self.when_date_string_is_parsed(date_string)
        self.then_detected_locale(locale)

    def test_try_previous_locales_false(self):
        self.given_parser(try_previous_locales=False)
        self.when_date_string_is_parsed("Mañana")  # es
        self.then_detected_locale("es")
        self.when_date_string_is_parsed("2020-05-01")
        self.then_detected_locale("en")

    def test_try_previous_locales_true(self):
        self.given_parser(try_previous_locales=True)
        self.when_date_string_is_parsed("Mañana")  # es
        self.then_detected_locale("es")
        self.when_date_string_is_parsed("2020-05-01")
        self.then_detected_locale("es")

    def test_try_previous_locales_order_deterministic(self):
        self.given_parser(try_previous_locales=True)
        self.when_date_string_is_parsed("Mañana")  # es
        self.then_detected_locale("es")
        self.when_date_string_is_parsed("Понедельник")  # ru
        self.then_detected_locale("ru")
        self.then_previous_locales_is(["es", "ru"])

    @parameterized.expand(
        [
            param("2014-10-09T17:57:39+00:00"),
        ]
    )
    def test_get_date_data_should_not_strip_timezone_info(self, date_string):
        self.given_parser()
        self.when_date_string_is_parsed(date_string)
        self.then_date_was_parsed()
        self.then_parsed_date_has_timezone()

    @parameterized.expand(
        [
            param(
                date_string="14 giu 13",
                date_formats=["%y %B %d"],
                expected_result=datetime(2014, 6, 13),
            ),
            param(
                date_string="14_luglio_15",
                date_formats=["%y_%B_%d"],
                expected_result=datetime(2014, 7, 15),
            ),
            param(
                date_string="14_LUGLIO_15",
                date_formats=["%y_%B_%d"],
                expected_result=datetime(2014, 7, 15),
            ),
            param(
                date_string="10.01.2016, 20:35",
                date_formats=["%d.%m.%Y, %H:%M"],
                expected_result=datetime(2016, 1, 10, 20, 35),
            ),
        ]
    )
    def test_parse_date_using_format(self, date_string, date_formats, expected_result):
        self.given_local_tz_offset(0)
        self.given_parser()
        self.when_date_string_is_parsed(date_string, date_formats)
        self.then_date_was_parsed()
        self.then_period_is("day")
        self.then_parsed_datetime_is(expected_result)

    @parameterized.expand(
        [
            param(
                date_string="11/09/2007", date_formats={"date_formats": ["%d/%m/%Y"]}
            ),
            param(date_string="16.09.03 11:55", date_formats=111),
            param(date_string="08-01-1998", date_formats=12.56),
        ]
    )
    def test_parsing_date_using_invalid_type_date_format_must_raise_error(
        self, date_string, date_formats
    ):
        self.given_local_tz_offset(0)
        self.given_parser()
        self.when_date_string_is_parsed(date_string, date_formats)
        self.then_error_was_raised(
            TypeError,
            [
                "Date formats should be list, tuple or set of strings",
                "'{}' object is not iterable".format(type(date_formats).__name__),
            ],
        )

    @parameterized.expand(
        [
            param(date_string={"date": "12/11/1998"}),
            param(date_string=[2017, 12, 1]),
            param(date_string=2018),
            param(date_string=12.2000),
            param(date_string=datetime(year=2009, month=12, day=7)),
        ]
    )
    def test_parsing_date_using_invalid_type_date_string_must_raise_error(
        self, date_string
    ):
        self.given_parser()
        self.when_date_string_is_parsed(date_string)
        self.then_error_was_raised(TypeError, ["Input type must be str"])

    @parameterized.expand(
        [
            param(
                date_string="2014/11/17 14:56 EDT",
                expected_result=datetime(2014, 11, 17, 18, 56),
            ),
        ]
    )
    def test_parse_date_with_timezones_not_using_formats(
        self, date_string, expected_result
    ):
        self.given_parser(settings={"TO_TIMEZONE": "UTC"})
        self.when_date_string_is_parsed(date_string)
        self.then_date_was_parsed()
        self.then_period_is("day")
        self.result["date_obj"] = self.result["date_obj"].replace(tzinfo=None)
        self.then_parsed_datetime_is(expected_result)

    @parameterized.expand(
        [
            param(
                date_string="2014/11/17 14:56 EDT",
                date_formats=["%Y/%m/%d %H:%M EDT"],
                expected_result=datetime(2014, 11, 17, 14, 56),
            ),
        ]
    )
    def test_parse_date_with_timezones_using_formats_ignore_timezone(
        self, date_string, date_formats, expected_result
    ):
        self.given_local_tz_offset(0)
        self.given_parser()
        self.when_date_string_is_parsed(date_string, date_formats)
        self.then_date_was_parsed()
        self.then_period_is("day")
        self.then_parsed_datetime_is(expected_result)

    @parameterized.expand(
        [
            param(
                date_string="08-08-2014\xa018:29",
                expected_result=datetime(2014, 8, 8, 18, 29),
            ),
        ]
    )
    def test_should_parse_with_no_break_space_in_dates(
        self, date_string, expected_result
    ):
        self.given_parser()
        self.when_date_string_is_parsed(date_string)
        self.then_date_was_parsed()
        self.then_period_is("day")
        self.then_parsed_datetime_is(expected_result)

    @parameterized.expand(
        [
            param(
                date_string="12 jan 1876",
                expected_result=(datetime(1876, 1, 12, 0, 0), "day", "en"),
            ),
            param(
                date_string="02/09/16",
                expected_result=(datetime(2016, 2, 9, 0, 0), "day", "en"),
            ),
            param(
                date_string="10 giu 2018",
                expected_result=(datetime(2018, 6, 10, 0, 0), "day", "it"),
            ),
        ]
    )
    def test_get_date_tuple(self, date_string, expected_result):
        self.given_parser()
        self.when_get_date_tuple_is_called(date_string)
        self.then_returned_tuple_is(expected_result)

    def given_now(self, year, month, day, **time):
        now = datetime(year, month, day, **time)
        datetime_mock = Mock(wraps=datetime)
        datetime_mock.utcnow = Mock(return_value=now)
        datetime_mock.now = Mock(return_value=now)
        datetime_mock.today = Mock(return_value=now)
        self.add_patch(patch("dateparser.date.datetime", new=datetime_mock))

    def given_parser(self, restrict_to_languages=None, **params):
        self.parser = date.DateDataParser(languages=restrict_to_languages, **params)

    def given_local_tz_offset(self, offset):
        self.add_patch(
            patch.object(
                dateparser.timezone_parser,
                "local_tz_offset",
                new=timedelta(seconds=3600 * offset),
            )
        )

    def when_date_string_is_parsed(self, date_string, date_formats=None):
        try:
            self.result = self.parser.get_date_data(date_string, date_formats)
        except Exception as error:
            self.error = error

    def when_multiple_dates_are_parsed(self, date_strings):
        self.multiple_results = []
        for date_string in date_strings:
            try:
                result = self.parser.get_date_data(date_string)
            except Exception as error:
                result = error
            finally:
                self.multiple_results.append(result)

    def when_get_date_tuple_is_called(self, date_string):
        self.result = self.parser.get_date_tuple(date_string)

    def then_date_was_parsed(self):
        self.assertIsNotNone(self.result["date_obj"])

    def then_date_locale(self):
        self.assertIsNotNone(self.result["locale"])

    def then_date_is_n_days_ago(self, days):
        today = datetime.now().date()
        expected_date = today - timedelta(days=days)
        self.assertEqual(expected_date, self.result["date_obj"].date())

    def then_all_results_were_parsed(self):
        self.assertNotIn(None, self.multiple_results)

    def then_parsed_dates_are(self, expected_dates):
        self.assertEqual(
            expected_dates,
            [result["date_obj"].date() for result in self.multiple_results],
        )

    def then_detected_locale(self, locale):
        self.assertEqual(locale, self.result["locale"])

    def then_period_is(self, day):
        self.assertEqual(day, self.result["period"])

    def then_parsed_datetime_is(self, expected_datetime):
        self.assertEqual(expected_datetime, self.result["date_obj"])

    def then_parsed_date_is(self, expected_date):
        self.assertEqual(expected_date, self.result["date_obj"].date())

    def then_parsed_date_has_timezone(self):
        self.assertTrue(hasattr(self.result["date_obj"], "tzinfo"))

    def then_returned_tuple_is(self, expected_tuple):
        self.assertEqual(expected_tuple, self.result)

    def then_previous_locales_is(self, expected_list):
        self.assertEqual(
            expected_list, [locale.shortname for locale in self.parser.previous_locales]
        )


class TestParserInitialization(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.parser = NotImplemented

    @parameterized.expand(
        [
            param(languages="en"),
            param(languages={"languages": ["en", "he", "it"]}),
        ]
    )
    def test_error_raised_for_invalid_languages_argument(self, languages):
        self.when_parser_is_initialized(languages=languages)
        self.then_error_was_raised(
            TypeError,
            ["languages argument must be a list (%r given)" % type(languages)],
        )

    @parameterized.expand(
        [
            param(locales="en-001"),
            param(locales={"locales": ["zh-Hant-HK", "ha-NE", "se-SE"]}),
        ]
    )
    def test_error_raised_for_invalid_locales_argument(self, locales):
        self.when_parser_is_initialized(locales=locales)
        self.then_error_was_raised(
            TypeError, ["locales argument must be a list (%r given)" % type(locales)]
        )

    @parameterized.expand(
        [
            param(region=["AW", "BE"]),
            param(region=150),
        ]
    )
    def test_error_raised_for_invalid_region_argument(self, region):
        self.when_parser_is_initialized(region=region)
        self.then_error_was_raised(
            TypeError, ["region argument must be str (%r given)" % type(region)]
        )

    @parameterized.expand(
        [
            param(try_previous_locales=["ar-OM", "pt-PT", "fr-CG", "uk"]),
            param(try_previous_locales="uk"),
            param(try_previous_locales={"try_previous_locales": True}),
            param(try_previous_locales=0),
        ]
    )
    def test_error_raised_for_invalid_try_previous_locales_argument(
        self, try_previous_locales
    ):
        self.when_parser_is_initialized(try_previous_locales=try_previous_locales)
        self.then_error_was_raised(
            TypeError,
            [
                "try_previous_locales argument must be a boolean (%r given)"
                % type(try_previous_locales)
            ],
        )

    @parameterized.expand(
        [
            param(use_given_order=["da", "pt", "ja", "sv"]),
            param(use_given_order="uk"),
            param(use_given_order={"use_given_order": True}),
            param(use_given_order=1),
        ]
    )
    def test_error_raised_for_invalid_use_given_order_argument(self, use_given_order):
        self.when_parser_is_initialized(
            locales=["en", "es"], use_given_order=use_given_order
        )
        self.then_error_was_raised(
            TypeError,
            [
                "use_given_order argument must be a boolean (%r given)"
                % type(use_given_order)
            ],
        )

    def test_error_is_raised_when_use_given_order_is_True_and_locales_and_languages_is_None(
        self,
    ):
        self.when_parser_is_initialized(use_given_order=True)
        self.then_error_was_raised(
            ValueError,
            ["locales or languages must be given if use_given_order is True"],
        )

    def test_no_error_for_order_with_languages_without_locales(self):
        self.when_parser_is_initialized(languages=["en", "fr"], use_given_order=True)

    def when_parser_is_initialized(
        self,
        languages=None,
        locales=None,
        region=None,
        try_previous_locales=True,
        use_given_order=False,
    ):
        try:
            self.parser = date.DateDataParser(
                languages=languages,
                locales=locales,
                region=region,
                try_previous_locales=try_previous_locales,
                use_given_order=use_given_order,
            )
        except Exception as error:
            self.error = error


class TestSanitizeDate(BaseTestCase):
    def test_remove_year_in_russian(self):
        self.assertEqual(date.sanitize_date("2005г."), "2005")
        self.assertEqual(date.sanitize_date("2005г. 20:37"), "2005 20:37")
        self.assertEqual(date.sanitize_date("2005 г."), "2005")
        self.assertEqual(date.sanitize_date("2005 г. 15:24"), "2005 15:24")
        self.assertEqual(date.sanitize_date("Авг."), "Авг")

    def test_sanitize_date_colons(self):
        self.assertEqual(date.sanitize_date("2019:"), "2019")
        self.assertEqual(date.sanitize_date("31/07/2019:"), "31/07/2019")


class TestDateLocaleParser(BaseTestCase):
    def setUp(self):
        super().setUp()

    @parameterized.expand(
        [
            param(
                date_data=DateData(date_obj=datetime(1999, 10, 1, 0, 0))
            ),  # missing period
            param(date_data=DateData(period="day")),  # missing date_obj
            param(
                date_data={"date": datetime(2007, 1, 22, 0, 0), "period": "day"}
            ),  # wrong format (dict)
            param(
                date_data=DateData(date_obj=datetime(2020, 10, 29, 0, 0), period="hour")
            ),  # wrong period
            param(
                date_data=[datetime(2007, 1, 22, 0, 0), "day"]
            ),  # wrong format (list)
            param(date_data=DateData(date_obj=None, period="day")),  # date_obj is None
            param(
                date_data=DateData(date_obj="29/05/1994", period="day")
            ),  # wrong date_obj format
        ]
    )
    def test_is_valid_date_data(self, date_data):
        self.given_parser(
            language=["en"],
            date_string="10 jan 2000",
            date_formats=None,
            settings=Settings(),
        )
        self.when_date_object_is_validated(date_data)
        self.then_date_object_is_invalid()

    def given_parser(self, language, date_string, date_formats, settings):
        self.parser = date._DateLocaleParser(
            language, date_string, date_formats, settings
        )

    def when_date_object_is_validated(self, date_data):
        self.is_valid_date_obj = self.parser._is_valid_date_data(date_data)

    def then_date_object_is_invalid(self):
        self.assertFalse(self.is_valid_date_obj)


class TestTimestampParser(BaseTestCase):
    def given_parser(self, **params):
        self.parser = date.DateDataParser(**params)

    def given_tzstr(self, tzstr):
        # Save the existing value
        self.old_tzstr = os.environ["TZ"] if "TZ" in os.environ else None

        # Overwrite the value, or remove it
        if tzstr is not None:
            os.environ["TZ"] = tzstr
        elif "TZ" in os.environ:
            del os.environ["TZ"]

        # Call tzset
        tzset()

    def reset_tzstr(self):
        # If we never set it with given_tzstr, don't bother resetting it
        if not hasattr(self, "old_tzstr"):
            return

        # Restore the old value, or remove it if null
        if self.old_tzstr is not None:
            os.environ["TZ"] = self.old_tzstr
        elif "TZ" in os.environ:
            del os.environ["TZ"]

        # Remove the local attribute
        del self.old_tzstr

        # Restore the old timezone behavior
        tzset()

    def test_timestamp_in_milliseconds(self):
        self.assertEqual(
            date.get_date_from_timestamp("1570308760263", None),
            datetime.fromtimestamp(1570308760).replace(microsecond=263000),
        )

    @parameterized.expand(
        product(
            ["1570308760"],
            ["EDT", "EST", "PDT", "PST", "UTC", "local"],
            [None, "EDT", "EST", "PDT", "PST", "UTC"],
            ["EST5EDT4", "UTC0", "PST8PDT7", None],
        )
    )
    def test_timestamp_with_different_timestr(
        self, timestamp, timezone, to_timezone, tzstr
    ):
        settings = {
            "RETURN_AS_TIMEZONE_AWARE": True,
            "TIMEZONE": timezone,
        }

        # is TO_TIMEZONE supposed to be allowed to be False, or None ???
        if to_timezone is not None:
            settings["TO_TIMEZONE"] = to_timezone

        self.given_parser(settings=settings)

        self.given_tzstr(tzstr)

        self.assertEqual(
            self.parser.get_date_data(timestamp)["date_obj"],
            datetime.fromtimestamp(int(timestamp), dttz.utc),
        )

        self.reset_tzstr()

    def test_timestamp_in_microseconds(self):
        self.assertEqual(
            date.get_date_from_timestamp("1570308760263111", None),
            datetime.fromtimestamp(1570308760).replace(microsecond=263111),
        )

    @parameterized.expand(
        [
            param(
                input_timestamp="-1570308760",
                negative=True,
                result=datetime.fromtimestamp(-1570308760),
            ),
            param(input_timestamp="-1570308760", negative=False, result=None),
            param(input_timestamp="1570308760", negative=True, result=None),
            param(
                input_timestamp="1570308760",
                negative=False,
                result=datetime.fromtimestamp(1570308760),
            ),
        ]
    )
    def test_timestamp_with_negative(self, input_timestamp, negative, result):
        self.assertEqual(
            date.get_date_from_timestamp(input_timestamp, None, negative=negative),
            result,
        )

    @parameterized.expand(
        [
            param(date_string="15703087602631"),
            param(date_string="157030876026xx"),
            param(date_string="1570308760263x"),
            param(date_string="157030876026311"),
            param(date_string="15703087602631x"),
            param(date_string="15703087602631xx"),
            param(date_string="15703087602631111"),
            param(date_string="1570308760263111x"),
            param(date_string="1570308760263111xx"),
            param(date_string="1570308760263111222"),
        ]
    )
    def test_timestamp_with_wrong_length(self, date_string):
        self.assertEqual(date.get_date_from_timestamp(date_string, None), None)


if __name__ == "__main__":
    unittest.main()



def pytest_collection_modifyitems(session, config, items):
    # Avoid executing tests when executing `--flake8` flag (pytest-flake8)
    try:
        from pytest_flake8 import Flake8Item

        if config.getoption("--flake8"):
            items[:] = [item for item in items if isinstance(item, Flake8Item)]
    except ImportError:
        pass
